(*Based on RifleLight*)
Require Export CTXT.

Section JoinLang.
Variable Z:Type.
Variable zvar_eq: forall (x y: Z), {x=y} + {x <> y}.  

Inductive JCom :=
  JSkip : JCom
| JAssign: PVar -> Expr -> JCom
| JComp: JCom -> JCom -> JCom
| JIte: Expr -> JCom -> JCom -> JCom
| JWhile: Expr -> JCom -> JCom
| JJoin: Z -> list Expr -> list Z -> JCom.

Fixpoint JErase (C:JCom) : Com :=
match C with
  JSkip => Skip
| JAssign x E => Assign x E
| JComp C1 C2 => Comp (JErase C1) (JErase C2)
| JIte e C1 C2 => Ite e (JErase C1) (JErase C2)
| JWhile e C => While e (JErase C)
| JJoin _ _ _ => Skip
end.

Definition zupd {A:Type} 
  (X:Z -> A) (z:Z) (a:A) : Z -> A :=
fun y => if zvar_eq z y then a else X y.

Definition ZVAR_MAP {A:Type}: MAP Z zvar_eq A zupd.
  constructor. unfold zupd. trivial. Defined.
Existing Instance ZVAR_MAP.

Lemma zvar_eq_ElimT: forall x p (H:(if zvar_eq p x then true else false) = true), p=x.
Proof.
 intros.
 remember (zvar_eq p x) as b; destruct b; subst; trivial. inversion H.
Qed.  
Hint Resolve zvar_eq_ElimT: varREWRITE. 

Lemma zvar_eq_ElimF: forall x p (H:(if zvar_eq p x then true else false) = false), p<>x.
Proof.
 intros.
 remember (zvar_eq p x) as b; destruct b; subst; trivial. inversion H.
Qed.  
Hint Resolve zvar_eq_ElimF: varREWRITE. 

Lemma zupd_split_sqsub: forall M M' x v z v'
(Case1: x = z -> ssqsub v v')
(Case2: x <> z -> ssqsub (M z) (M' z)),
ssqsub (zupd M x v z) (zupd M' x v' z).
Proof.
intros.
unfold zupd.
remember (zvar_eq x z) as b; destruct b; subst;  auto.
Qed.

(*making parameter A implicit here would prevent me from 
  adding the lemma to the hint database polymorphically*)
Lemma zvar_eq_refl: forall A z (a b:A),
  (if zvar_eq z z then a else b) = a.
Proof. intros.
  remember (zvar_eq z z) as bb; destruct bb; subst; trivial. 
  apply False_ind.  apply n.  trivial.
Qed. 
Hint Rewrite zvar_eq_refl: varREWRITE.

Definition ZStore : Type := Z -> L. 

Definition zeval (M:ZStore) (xs: list Z) :=
fold_right (fun z => fun l => llub (M z) l) bbot xs. 

Lemma zeval_app: forall M l1 l2, zeval M (l1 ++ l2) =
llub (zeval M l1) (zeval M l2).
Proof.
intros.
induction l1; unfold zeval in *; simpl. rewrite bbot_id. trivial.
rewrite IHl1. rewrite llub_assoc. trivial.
Qed.

Lemma zeval_CONS: forall M z K, 
zeval M (z::K) = llub (M z) (zeval M K).
Proof.
intros.
reflexivity.
Qed.

Lemma zeval_IN:
  forall K z M,
  In z K -> ssqsub (M z) (zeval M K).
Proof.
intros K.
induction K; intros; simpl in *. 
  contradiction. 
destruct H. subst. unfold zeval; simpl. apply lub_leqleft.
  eapply leq_trans. apply (IHK _ _ H).
  unfold zeval; simpl. apply lub_leqright.
Qed.

Lemma zeval_SUBLIST: forall M K K',
 (forall z, In z K -> In z K') ->
 ssqsub (zeval M K) (zeval M K').
Proof.
intros M K.
induction K; simpl; intros.
  unfold zeval in *; simpl. apply bot_leq.
apply lub_leq. apply zeval_IN. apply H. left; trivial. 
  apply IHK. intros. apply H. right; assumption.
Qed.

Definition lab_eval (G:CTXT) (es: list Expr) :=
fold_right llub bbot (map (GEval G) es). 

Lemma lab_eval_GEval : forall G e, (lab_eval G (e :: nil)) = GEval G e.
Proof.
intros.
unfold lab_eval; simpl.
apply id_bbot.
Qed.   

Definition JStore:Type := prod (prod Store CTXT) ZStore. 

Inductive JoinSem: JStore -> JCom -> JStore -> Prop :=
  JS_Skip: forall s, JoinSem s JSkip s
| JS_Assign: forall x e s G t D M, 
     t = upd s x (eval s e) ->
     D = upd G x (GEval G e) -> 
     JoinSem (s,G,M) (JAssign x e) (t,D,M)
| JS_Comp: forall s C1 C2 t r, 
     JoinSem s C1 r -> JoinSem r C2 t -> JoinSem s (JComp C1 C2) t
| JS_True: forall s E C1 C2 t, istrue (eval (fst (fst s)) E) = true ->
           JoinSem s C1 t -> JoinSem s (JIte E C1 C2) t
| JS_False: forall s E C1 C2 t, istrue (eval (fst (fst s)) E) = false -> 
              JoinSem s C2 t -> JoinSem s (JIte E C1 C2) t
| JS_WTrue: forall s b C r t, istrue (eval (fst (fst s)) b) = true -> 
           JoinSem s C r -> JoinSem r (JWhile b C) t -> 
           JoinSem s (JWhile b C) t
| JS_WFalse: forall s b C t, istrue (eval (fst (fst s)) b) = false -> s=t -> 
           JoinSem s (JWhile b C) t
| JS_Join : forall s G M z N zs es, 
            N = zupd M z (llub (lab_eval G es) (zeval M zs)) ->             
            JoinSem (s,G,M) (JJoin z es zs) (s,G,N).  

Lemma JErase_Lemma: forall s R t (Hyp:JoinSem s R t),
Sem (fst (fst s)) (JErase R) (fst (fst t)). 
Proof. 
intros s R t Hyp. 
induction Hyp; intros; try subst; simpl in *.
(*Skip*) 
  constructor.  trivial. 
(*Assign*) 
  constructor.  trivial. 
(*Comp*)
  econstructor. apply IHHyp1. apply IHHyp2. 
(*IfTrue*)
    apply S_True. assumption. apply IHHyp.
(*IfFalse*)
    apply S_False. assumption. apply IHHyp.
(*WhileTrue*)
    eapply S_WTrue. assumption. apply IHHyp1. apply IHHyp2.
(*WhileFalse*)
    apply S_WFalse. assumption. trivial.
(*Join*)
   constructor. trivial.
Qed.

Lemma JErase_Lemma_Inv: forall R s t (Hyp:Sem s (JErase R) t) G M,
  exists H, exists N, JoinSem (s,G,M) R (t,H,N). 
Proof. 
intros R.
induction R. 
(*Skip*) intros.
  inversion Hyp; subst; simpl; clear Hyp. 
  eexists. eexists.
  constructor.
(*Assign *) intros.
  inversion Hyp; subst; simpl; clear Hyp. 
  eexists; eexists.
  econstructor.  reflexivity.   reflexivity.
(*Comp*) intros.
  inversion Hyp; subst; simpl; clear Hyp. 
  destruct (IHR1 _ _ H2 G M) as [H [P HR]]. 
  destruct (IHR2 _ _ H4 H P) as [K [N HR2]]. 
  eexists; eexists.
  econstructor.  apply HR.  apply HR2. 
(*If*) intros.
  inversion Hyp; subst; simpl; clear Hyp. 
  (*True*) clear IHR2. 
     destruct (IHR1 _ _ H5 G M) as [H [N HR]]. 
     eexists; eexists.
     apply JS_True.  apply H4.  apply HR.  
  (*False*) clear IHR1. 
     destruct (IHR2 _ _ H5 G M) as [H [N HR]]. 
     eexists; eexists.
     apply JS_False.  apply H4.  apply HR.
(*While*) 
  intros s t SEM. simpl in SEM.
  remember (While e (JErase R)) as D.
  induction SEM; subst; try inversion HeqD; subst; clear HeqD; intros.
  (*WhileT*)
      destruct (IHR _ _ SEM1 G M) as [K [P Kv]].
      destruct (IHSEM2 (eq_refl _) K P) as [L [N Lw]].
      exists L. exists N. eapply JS_WTrue. apply H. apply Kv. apply Lw.
  (*WhileT*)
      exists G. exists M. eapply JS_WFalse. apply H. trivial.
(*Join*)
  intros.
  inversion Hyp; subst; simpl; clear Hyp. 
  eexists; eexists. constructor. reflexivity.
Qed. 

Theorem JoinJErase: forall R s t G M,
Sem s (JErase R) t <-> (exists D, exists N, JoinSem (s,G,M) R (t,D,N)). 
Proof.
intros.
split; intros.
  apply JErase_Lemma_Inv. apply H.
destruct H as [D [E HH]].
  apply (JErase_Lemma _ _ _ HH).
Qed.

Theorem ControlTaints_Dont_Influence_DataTaints:
forall J s G M t D N (Hyp: JoinSem (s,G,M) J (t,D,N))
       M' t' D' N' (Hyp': JoinSem (s,G,M') J (t',D',N')),
t=t' /\ D=D'.
Proof.
intros J.
induction J; intros; subst; simpl in*. 
(*Skip*)
  inversion Hyp; subst; simpl; clear Hyp. 
  inversion Hyp'; subst; simpl; clear Hyp'. 
  split; trivial.
(*Assign *) 
  inversion Hyp; subst; simpl; clear Hyp. 
  inversion Hyp'; subst; simpl; clear Hyp'. 
  split; trivial.
(*Comp*) 
  inversion Hyp; subst; simpl; clear Hyp. destruct r as [[r K] P].
  inversion Hyp'; subst; simpl; clear Hyp'. destruct r0 as [[r' K'] P']. 
  destruct (IHJ1 _ _ _ _ _ _ H2 _ _ _ _ H3) as [Hrr' HKK']; clear IHJ1 H2 H3.
  subst.
  apply (IHJ2 _ _ _ _ _ _ H4 _ _ _ _ H6).
(*If*) 
  inversion Hyp; subst; simpl; clear Hyp; simpl in *. 
  (*True*)
    inversion Hyp'; subst; simpl; clear Hyp'; simpl in *.
    (*True*)
      apply (IHJ1 _ _ _ _ _ _ H5 _ _ _ _ H7).
    (*False*)
      rewrite H6 in H4. inversion H4.   
  (*True*)
    inversion Hyp'; subst; simpl; clear Hyp'; simpl in *.
    (*True*)
      rewrite H6 in H4. inversion H4. 
    (*False*)
      apply (IHJ2 _ _ _ _ _ _ H5 _ _ _ _ H7).
(*While*) 
  remember (s,G,M) as ST. remember (t,D,N) as TT.
  generalize dependent N'. generalize dependent D'. generalize dependent t'. generalize dependent M'. 
  generalize dependent N. generalize dependent D. generalize dependent t.
  generalize dependent M. generalize dependent G. generalize dependent s.
  remember (JWhile e J) as LOOP.
  induction Hyp; try inversion HeqLOOP; clear HeqLOOP.
  (*True*)
     intros; subst. clear IHHyp1 Hyp2; simpl in *. rename s0 into s. rename t0 into t.
     destruct r as [[r K] P].
     assert (IH:= IHHyp2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _)). clear IHHyp2.
     assert (IJ := IHJ _ _ _ _ _ _ Hyp1). clear Hyp1 IHJ.
     inversion Hyp'; subst; clear Hyp'; simpl in *.
     (*True*)
        destruct r0 as [[r' K'] P'].
        destruct (IJ _ _ _ _ H4) as [Hrr' HKK']; subst; clear IJ H4.
        apply (IH _ _ _ _ H6).
     (*False*)
        inversion H5; subst; clear H5.
        rewrite H3 in H; inversion H.
  (*False*)
     intros; subst. clear IHJ; simpl in *. rename s0 into s. rename t0 into t.
     inversion HeqTT; subst; clear HeqTT.
     inversion Hyp'; subst; clear Hyp'; simpl in *.
     (*True*)
        rewrite H2 in H; inversion H.
     (*False*)
        inversion H5; subst; clear H5.
        split; trivial.
(*Join*)
  inversion Hyp; subst; simpl; clear Hyp. 
  inversion Hyp'; subst; simpl; clear Hyp'. 
  split; trivial.
Qed.


Theorem Taints_Dont_Influence_DataPlane:
forall J s G M t D N (Hyp: JoinSem (s,G,M) J (t,D,N))
         G' M' t' D' N' (Hyp': JoinSem (s,G',M') J (t',D',N')),
t=t'.
Proof.
intros J.
induction J; intros; subst; simpl in*. 
(*Skip*)
  inversion Hyp; subst; simpl; clear Hyp. 
  inversion Hyp'; subst; simpl; clear Hyp'. 
  trivial.
(*Assign *) 
  inversion Hyp; subst; simpl; clear Hyp. 
  inversion Hyp'; subst; simpl; clear Hyp'. 
  trivial.
(*Comp*) 
  inversion Hyp; subst; simpl; clear Hyp. destruct r as [[r K] P].
  inversion Hyp'; subst; simpl; clear Hyp'. destruct r0 as [[r' K'] P']. 
  rewrite (IHJ1 _ _ _ _ _ _ H2 _ _ _ _ _ H3) in *; clear IHJ1 H2 H3.
  apply (IHJ2 _ _ _ _ _ _ H4 _ _ _ _ _ H6).
(*If*) 
  inversion Hyp; subst; simpl; clear Hyp; simpl in *. 
  (*True*)
    inversion Hyp'; subst; simpl; clear Hyp'; simpl in *.
    (*True*)
      apply (IHJ1 _ _ _ _ _ _ H5 _ _ _ _ _ H7).
    (*False*)
      rewrite H6 in H4. inversion H4.   
  (*True*)
    inversion Hyp'; subst; simpl; clear Hyp'; simpl in *.
    (*True*)
      rewrite H6 in H4. inversion H4. 
    (*False*)
      apply (IHJ2 _ _ _ _ _ _ H5 _ _ _ _ _ H7).
(*While*) 
  remember (s,G,M) as ST. remember (t,D,N) as TT.
  generalize dependent N'. generalize dependent D'. generalize dependent t'.
  generalize dependent M'. generalize dependent G'. 
  generalize dependent N. generalize dependent D. generalize dependent t.
  generalize dependent M. generalize dependent G. generalize dependent s.
  remember (JWhile e J) as LOOP.
  induction Hyp; try inversion HeqLOOP; clear HeqLOOP.
  (*True*)
     intros; subst. clear IHHyp1 Hyp2; simpl in *. rename s0 into s. rename t0 into t.
     destruct r as [[r K] P].
     assert (IH:= IHHyp2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _)). clear IHHyp2.
     assert (IJ := IHJ _ _ _ _ _ _ Hyp1). clear Hyp1 IHJ.
     inversion Hyp'; subst; clear Hyp'; simpl in *.
     (*True*)
        destruct r0 as [[r' K'] P'].
        rewrite (IJ _ _ _ _ _ H4) in * ; clear IJ H4.
        apply (IH _ _ _ _ _ H6).
     (*False*)
        inversion H5; subst; clear H5.
        rewrite H3 in H; inversion H.
  (*False*)
     intros; subst. clear IHJ; simpl in *. rename s0 into s. rename t0 into t.
     inversion HeqTT; subst; clear HeqTT.
     inversion Hyp'; subst; clear Hyp'; simpl in *.
     (*True*)
        rewrite H2 in H; inversion H.
     (*False*)
        inversion H5; subst; clear H5.
        trivial.
(*Join*)
  inversion Hyp; subst; simpl; clear Hyp. 
  inversion Hyp'; subst; simpl; clear Hyp'. 
  trivial.
Qed.

Lemma JoinSem_nomod: forall s R t (SEM:JoinSem s R t) x 
 (MV:modVars (JErase R) x = false), fst (fst s) x = fst (fst t) x.
Proof.
intros s R t SEM.
induction SEM; intros; subst; simpl in *; trivial; simpl in MV. 
(*Assign*)
apply pvar_eq_ElimF in MV.
rewrite M_upd_Diff; trivial.
(*Comp*)
apply Bool.orb_false_iff in MV. destruct MV as [MV1 MV2].
rewrite <- (IHSEM2 x MV2).
apply (IHSEM1 _ MV1).
(*If-True*)
apply Bool.orb_false_iff in MV. destruct MV as [MV1 MV2].
apply (IHSEM x MV1).
(*If-False*)
apply Bool.orb_false_iff in MV. destruct MV as [MV1 MV2].
apply (IHSEM x MV2).
(*While*)
rewrite <- (IHSEM2 x MV).
apply (IHSEM1 _ MV).
Qed. 

Lemma JSem1_aux: forall R pre post
(Hyp: JoinSem pre R post) x,
(modVars (JErase R) x = false) ->
((fst (fst pre)) x = (fst (fst post)) x /\ snd(fst pre) x = snd (fst post) x). 
Proof. 
intros R pre post Hyp. 
induction Hyp; subst; intros; simpl in *. 
(*Skip*)
 split; trivial.
(*Assign*)  
  apply pvar_eq_ElimF in H.
  repeat rewrite M_upd_Diff; trivial. split; trivial.
(*Comp*)
  apply Bool.orb_false_iff in H. destruct H as [MV1 MV2].
  destruct (IHHyp1 _ MV1) as [HH1 HH2].  rewrite HH1.  rewrite HH2. 
  apply (IHHyp2 _ MV2). 
(*True*)
  apply Bool.orb_false_iff in H0. destruct H0 as [MV1 MV2].
  apply (IHHyp _ MV1). 
(*False*)
  apply Bool.orb_false_iff in H0. destruct H0 as [MV1 MV2].
  apply (IHHyp _ MV2). 
(*WhileT*)
  intros. 
  destruct (IHHyp2 x H0). rewrite <- H1. rewrite <- H2. clear H1 H2.
  apply IHHyp1. apply H0.
(*WhileF*)
  split; trivial.
(*Join*)
  split; trivial. 
Qed. 

Lemma JSem1: forall R s G t H M N
(Hyp: JoinSem (s,G,M) R (t,H,N)) x,
(modVars (JErase R) x = false) -> (s x = t x /\ G x = H x). 
Proof. 
intros. 
apply (JSem1_aux _ _ _ Hyp _ H0). 
Qed.
End JoinLang.
