(*Based on mkUProg*)
Require Export JoinLanguage.
Require Export FPC.

Parameter fresh : forall (X : list nat), { n : nat | In n X -> False}.
Inductive SVar := 
  var_svar : Var -> SVar (*for each program variable x we have one security variable z^x.
                            Later, we;'ll also use z^pc*)
| cond_svar: nat -> SVar. (*for each i in N we have one plus a security variable z^i
                         -- used whenever we enter another branch*)

Lemma SVar_eq : forall (x y: SVar), {x=y} + {x <> y}.
Proof.
intros.
destruct x; destruct y.
    remember (var_eq v v0) as b. destruct b; subst. 
      left. trivial.
      right. intros K. apply n. inversion K. trivial.
    right. discriminate.
    right. discriminate.
    remember (EqNat.eq_nat_decide n n0) as b; destruct b.
      rewrite (EqNat.eq_nat_eq _ _ e). left. trivial.
      right. intros K. inversion K; subst. apply n1. apply EqNat.eq_nat_refl.
Qed.  

Lemma svar_eq_refl: forall A z (a b:A),
  (if SVar_eq z z then a else b) = a.
Proof. intros.
  remember (SVar_eq z z) as bb; destruct bb; subst; trivial. 
  apply False_ind.  apply n.  trivial.
Qed. 

Lemma svar_eq_diff:
  forall A (z x: SVar) (a b : A), 
  z <> x -> (if SVar_eq z x then a else b) = b.
Proof.
intros. remember (SVar_eq z x) as bb; destruct bb; subst; trivial. exfalso. apply H. trivial.
Qed.

(*amounts to z(e) written in the paper*)
Fixpoint mkSVars (E:Expr) : list SVar :=
match E with
  valExpr v => nil
| varExpr x => (var_svar (progVar x))::nil
| bopExpr e1 b e2 => (mkSVars e1) ++ (mkSVars e2)
end. 

Lemma zeval_GEval: forall M e,
zeval SVar M (mkSVars e) = GEval (fun x : PVar => M (var_svar (progVar x))) e.
intros.
induction e.
  trivial. 
  simpl. rewrite id_bbot. trivial.  
simpl. rewrite zeval_app.
  rewrite IHe1. rewrite IHe2. trivial.
Qed.

Definition supd := @zupd _ SVar_eq L.
Definition supd_Same := @M_upd_Same _ SVar_eq _ _ (@ZVAR_MAP _ SVar_eq L).
Definition supd_Diff := @M_upd_Diff _ SVar_eq _ _ (@ZVAR_MAP _ SVar_eq L).
Definition supd_split_at := @M_upd_split_at _ SVar_eq _ _ (@ZVAR_MAP _ SVar_eq L).

Lemma supd_split_sqsub: forall M M' x v z v'
(Case1: x = z -> ssqsub v v')
(Case2: x <> z -> ssqsub (M z) (M' z)),
ssqsub (supd M x v z) (supd M' x v' z).
Proof.
intros.
apply zupd_split_sqsub; assumption. 
Qed.

Lemma zeval_zupd: forall M n e kappa,
zeval SVar M (mkSVars e) = zeval SVar (supd M (cond_svar n) kappa) (mkSVars e).
Proof.
intros.
induction e.
  trivial.
  simpl. rewrite supd_Diff; try discriminate. trivial.
  simpl. repeat rewrite zeval_app. rewrite IHe1. rewrite IHe2. trivial.
Qed.

Lemma zeval_pc_supd: forall M e kappa,
zeval SVar M (mkSVars e) = zeval SVar (supd M (var_svar pcVar) kappa) (mkSVars e).
Proof.
intros.
induction e.
  trivial.
  simpl. rewrite supd_Diff; try discriminate. trivial.
  simpl. repeat rewrite zeval_app. rewrite IHe1. rewrite IHe2. trivial.
Qed.

Lemma zeval_IN_svar: forall L y M (HL: In y L), 
ssqsub (M y) (zeval SVar M L).
Proof.
intros.
apply (zeval_IN _ _ _ _ HL). 
Qed.

Definition mkJoin (n:nat) (x:PVar) : JCom SVar :=
   JJoin _ (var_svar (progVar x)) nil ((cond_svar n) ::(var_svar (progVar x)) :: nil).

Lemma mkJoin_exists: forall ST n x,  
exists TT, JoinSem SVar SVar_eq ST (mkJoin n x) TT.
Proof.
intros. destruct ST as [[s G] M]. eexists. constructor. reflexivity.
Qed.

(*Called Join in paper*)
Fixpoint mkJoins (n:nat) (ys:list PVar) : JCom SVar :=
  match ys with nil => JSkip SVar
              | (y :: yss) => JComp _ (mkJoin n y) (mkJoins n yss)
  end.

Lemma mkJoins_exists: forall Vars ST n, 
exists TT, JoinSem SVar SVar_eq ST (mkJoins n Vars ) TT.
Proof.
intros Vars.
induction Vars; intros; simpl.
  eexists. constructor.
destruct (mkJoin_exists ST n a) as [TT HTT]. (*[[[t D] N] H].*)
destruct (IHVars TT n) as [RR HR].
exists RR.
  econstructor. apply HTT. apply HR.
Qed.

Lemma JErase_mkJoins_Sem: forall Vars t n, Sem t (JErase SVar (mkJoins n Vars)) t.
Proof.
intros Vars.
induction Vars; intros; subst; simpl in *.
  constructor. reflexivity.
econstructor. 
  constructor. reflexivity.
  apply IHVars.
Qed.

Lemma mkJoin_effect: forall s G M (y:PVar) n t D N, 
JoinSem _ SVar_eq (s, G,M) (mkJoin n y) (t,D,N) -> 
t=s /\ D=G /\ N = supd M (var_svar (progVar y)) 
                             ((llub (M (cond_svar n))
                                                (M (var_svar (progVar y))))).
Proof.
intros.
inversion H. subst. clear H.
split; trivial.
split; trivial.
simpl. repeat rewrite bbot_id. rewrite id_bbot. trivial. 
Qed.

Lemma mkJoins_effect: forall n (Mods:list PVar) s G M t D N, 
JoinSem _ SVar_eq (s, G,M) (mkJoins n Mods) (t,D,N) -> 
t=s /\ D=G /\ N = (fun (y:SVar) => 
                     match y with
                       var_svar yy =>
                         match yy with
                           progVar yyy =>
                              if (@in_dec _ pvar_eq yyy Mods )
                              then (llub (M (cond_svar n))
                                      (M (var_svar (progVar yyy))))
                              else M (var_svar (progVar yyy))
                         | pcVar => M (var_svar pcVar)
                         end
                     | cond_svar m => M (cond_svar m)
                     end).
Proof.
intros n Mods.
induction Mods; intros; simpl in H.
  inversion H; subst; clear H.
  split; trivial.
  split; trivial.
  apply extensionality. intros x.
    destruct x; trivial.
    destruct v; simpl; trivial.
(*Cons*)
  inversion H; subst; clear H. destruct r as [[r E] P].
  destruct (mkJoin_effect _ _ _ _ _ _ _ _ H3) as [K1 [K2 K3]]. subst. clear H3.
  destruct (IHMods _ _ _ _ _ _ H5) as [K1 [K2 K3]]. subst. clear IHMods H5.
  split; trivial.
  split; trivial.
  apply extensionality. intros x.
    destruct x; simpl.
      destruct v; simpl. rewrite supd_Diff; try discriminate. trivial.
           rewrite supd_Diff; try discriminate.
           remember (in_dec pvar_eq p Mods) as b; destruct b; subst; simpl.
             remember (pvar_eq a p) as bb; destruct bb; subst; simpl in *.
               rewrite supd_Same. rewrite llub_assoc.  
                    rewrite lub_refl. trivial.
               rewrite supd_Diff; trivial. intros Z. inversion Z; subst. apply n0. trivial.
           apply supd_split_at; intros.
             inversion H; subst; clear H.
              remember (pvar_eq p p) as bb; destruct bb; subst; simpl in *. trivial.
              exfalso. apply n1. trivial.
           remember (pvar_eq a p) as bb; destruct bb; subst; simpl in *.
              exfalso. apply H. trivial.
            trivial.
    rewrite supd_Diff; try discriminate. trivial.
Qed.


Lemma modVars_mkJoin: forall n y, 
modVars (JErase SVar (mkJoin n y)) = fun z => false.
Proof.
intros. reflexivity.
Qed.

Lemma modVars_mkJoins: forall n Y, 
modVars (JErase SVar (mkJoins n Y)) = fun z => false.
Proof.
intros.
induction Y; simpl. reflexivity.
rewrite IHY. trivial.
Qed.

Lemma mkJoins_ControlLabelsPreserved: 
  forall V s G M t D N n
   (Hyp: JoinSem _ SVar_eq (s,G,M)
              (mkJoins n V)
              (t,D,N)) m,
   M (cond_svar m) = N (cond_svar m).
Proof. 
intros V. 
induction V; intros; simpl in *. 
inversion Hyp; subst; clear Hyp. trivial.
inversion Hyp; subst; clear Hyp. destruct r as [[r E] P].
    rewrite <- (IHV _ _ _ _ _ _ _ H4).
    clear IHV H4.
    inversion H2; subst; clear H2.
    rewrite supd_Diff; try discriminate. trivial.
Qed.

Lemma mkJoins_Notin_SecStore1: forall V s G M t D N n
   (Hyp: JoinSem _ SVar_eq (s,G,M)
              (mkJoins n V)
              (t,D,N))
    y (MV : ~In y V),
   M (var_svar (progVar y)) = N (var_svar (progVar y)). 
Proof. 
intros V. 
induction V; intros; simpl in *. 
inversion Hyp; subst; clear Hyp. trivial.
inversion Hyp; subst; clear Hyp. destruct r as [[r E] P].
    assert (HN: ~ In y V). intros KK. apply MV. right. apply KK.
    rewrite <- (IHV _ _ _ _ _ _ _ H4 _  HN).
    clear IHV H4 HN.
    inversion H2; subst; clear H2.
    simpl. rewrite id_bbot. rewrite bbot_id.
   apply eq_sym.
   apply supd_split_at; intros.
      inversion H; subst. exfalso. apply MV. left; trivial.
    trivial.
Qed.

Lemma mkJoins_Notin: forall V s G M t D N n  
    y (MV : ~In y V)
   (Hyp: JoinSem _ SVar_eq (s,G,M)
              (mkJoins n V)
              (t,D,N)),
   t y = s y /\ D y = G y /\ (forall z, ssqsub (M z) (N z)). 
Proof. (* with (autorewrite with LREW using try reflexivity).*)
intros V. 
induction V; intros; simpl in *. 
inversion Hyp; subst; clear Hyp. 
   split; trivial.  split; trivial.
    intros. apply leq_refl. 
inversion Hyp; subst; clear Hyp. destruct r as [[r E] P].
    assert (HN: ~ In y V). intros Z. apply MV. right. apply Z.
    destruct (IHV _ _ _ _ _ _ _ _ HN H4) as [K1 [K2 K3]].
    rewrite K1 in *; clear K1.
    rewrite K2 in *; clear K2.
    clear IHV H4.
    inversion H2; subst; clear H2.
    split. trivial. 
    split; trivial.
    intros.  
    eapply leq_snart. apply K3. 
    simpl. repeat rewrite bbot_id. repeat rewrite id_bbot.
    destruct z. 
      remember (var_eq v (progVar a)) as b; destruct b; subst; simpl in *.
        rewrite supd_Same. apply lub_leqright.
      rewrite supd_Diff; try discriminate. apply leq_refl. intros Z. inversion Z; subst. apply n0. trivial.
    rewrite supd_Diff; try discriminate. apply leq_refl.
Qed.

Lemma mkJoins_In: forall V s G M t D N n  
    y (MV : In y V)
   (Hyp: JoinSem _ SVar_eq (s,G,M)
              (mkJoins n V)
              (t,D,N)),
   ssqsub (M (cond_svar n)) (N (var_svar (progVar y))). 
Proof. 
intros V. 
induction V; intros; simpl in *. contradiction. 
inversion Hyp; subst; clear Hyp. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  remember (in_dec pvar_eq y V) as b; destruct b; subst; simpl.
    clear MV.
    assert (Z:= IHV _ _ _ _ _ _ _ _ i H4). clear IHV H4.
    rewrite supd_Diff in Z; try discriminate.
    apply Z.
  destruct MV; subst.
    clear IHV.
    assert (Z:= mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H4 _ n0).
    clear H4.
    rewrite supd_Same in Z.
    rewrite <- Z.
    apply lubright.  apply lub_leqleft.
  exfalso. apply n0. apply H.
Qed.


Lemma nNil: forall (n:nat), ~ In n nil.
Proof. 
intros. simpl. intros N. trivial. Qed. 

Definition topPC (F:list nat) : list SVar := 
match F with nil => nil
 | cons n t => (cond_svar n)::nil
end.

(*Amounts to z(F) in the paper*)
Function allPCs (F:list nat) : list SVar :=
match F with nil => nil
 | cons n t => (cond_svar n)::(allPCs t)
end.

Lemma allPCs_In: forall F n, In n F -> In (cond_svar n) (allPCs F).
Proof.
intros F.
induction F; intros; subst; simpl in *. assumption.
destruct H; subst.
  left; trivial.
  right. apply (IHF _ H).
Qed.

Lemma allPCs_InRev1: forall F n, In (cond_svar n) (allPCs F) -> In n F.
Proof.
intros F.
induction F; intros; subst; simpl in *. assumption.
destruct H.
  left. inversion H. trivial.
  right. apply (IHF _ H).
Qed.

Lemma allPCs_InRev2: forall F x, In x (allPCs F) -> 
exists n, x = (cond_svar n) /\ In n F.
Proof.
intros F.
induction F; intros; subst; simpl in *. contradiction.
destruct H; subst.
  exists a. split; trivial. left. trivial.
  destruct (IHF _ H) as [n [NN1 NN2]]. subst.
  exists n. split; trivial. right. trivial.
Qed.

Lemma allPCs_eq: forall M M' treg F
(HMM' : forall z : SVar, z <> cond_svar treg -> M z = M' z)
(NIN: ~ In treg F),
zeval SVar M (allPCs F) = zeval SVar M' (allPCs F).
Proof.
intros M M' treg F HMM'.
induction F; intros; subst.
  reflexivity.
simpl. repeat rewrite zeval_CONS. simpl.
  rewrite (HMM' (cond_svar a)).
    rewrite IHF. trivial.
    intros Z. apply NIN. right; trivial.
  intros Z. inversion Z; subst. apply NIN. left; trivial.
Qed.

Lemma allPCs_pcVar_eq:
  forall (M M' : SVar -> L) (F : list nat),
  (forall z : SVar, z <> var_svar pcVar -> M z = M' z) ->
  zeval SVar M (allPCs F) = zeval SVar M' (allPCs F).
Proof.
intros.
induction F; intros; subst; simpl.
  reflexivity.
repeat rewrite zeval_CONS.
 simpl. rewrite H; try discriminate.
 rewrite IHF. trivial.
Qed.

Lemma zeval_allPCs_supd: forall F M n k,
~ In n F ->
  zeval SVar M (allPCs F) = zeval SVar (supd M (cond_svar n) k) (allPCs F).
Proof.
intros.
eapply (allPCs_eq _ _ n).
intros. rewrite supd_Diff. trivial. intros Z. subst. apply H0. trivial.
assumption.
Qed.

Lemma zeval_allPCs_pc_supd: forall F M k,
  zeval SVar M (allPCs F) = zeval SVar (supd M (var_svar pcVar) k) (allPCs F).
Proof.
intros.
apply allPCs_pcVar_eq.
intros. rewrite supd_Diff. trivial. intros Z. subst. apply H. trivial.
Qed.

Lemma allPCs_upd_progVar: forall F M x p,
zeval SVar (supd M (var_svar (progVar x)) p) (allPCs F) =
zeval SVar M (allPCs F).
Proof.
intros F.
induction F; simpl; intros; subst.
reflexivity.
repeat rewrite zeval_CONS. rewrite IHF. simpl.
rewrite supd_Diff. trivial. discriminate.
Qed.

Lemma mkSvars_upd_cond: forall E M n p,
(zeval SVar (supd M (cond_svar n) p) (mkSVars E)) = (zeval SVar M (mkSVars E)).
Proof.
intros.
repeat rewrite zeval_GEval.
induction E; simpl; trivial.
rewrite supd_Diff; trivial. discriminate.
rewrite IHE1. rewrite IHE2. trivial.
Qed.

Lemma zeval_all_app: 
  forall F E M n p (HN: ~ In n F),
  (zeval SVar (supd M (cond_svar n) p) (allPCs F ++ mkSVars E)) =
  (zeval SVar M (allPCs F ++ mkSVars E)).
Proof.
intros. 
repeat rewrite zeval_app.
rewrite mkSvars_upd_cond.
apply lub_cong; trivial.
  rewrite <- zeval_allPCs_supd; trivial.
Qed.

Lemma zeval_app_pc_supd:
  forall (F : list nat) (M : ZStore SVar) (k : L) E,
  zeval SVar M (allPCs F ++ mkSVars E) =
  zeval SVar (supd M (var_svar pcVar) k) (allPCs F ++ mkSVars E).
Proof.
intros.
repeat rewrite zeval_app.
rewrite <- zeval_allPCs_pc_supd.
rewrite <- zeval_pc_supd.
trivial.
Qed.


Lemma GEval_supd_condsvar: forall E M n p, 
GEval (fun x => (supd M (cond_svar n) p) (var_svar (progVar x))) E = 
GEval (fun x => M (var_svar (progVar x))) E.
Proof.
intros. induction E; simpl; intros; subst.
trivial.
rewrite supd_Diff. trivial. discriminate.
rewrite IHE1. rewrite IHE2. trivial.
Qed.

Lemma allPCs_conds_eq: forall F P Q
(HF: forall n, In n F -> P (cond_svar n) = Q (cond_svar n)),
(zeval SVar P (allPCs F)) =  (zeval SVar Q (allPCs F)).
Proof.
intros.
induction F; simpl; trivial.
repeat rewrite zeval_CONS; simpl.
rewrite (HF a).
  rewrite IHF. trivial. intros. apply HF. right; trivial.
left; trivial.
Qed.

Lemma mkSVars_pcVar_eq:
  forall (M M': ZStore SVar) (e : Expr),
  (forall z : SVar, z <> var_svar pcVar -> M z = M' z) ->
  zeval SVar M (mkSVars e) =
  zeval SVar M' (mkSVars e).
Proof.
intros.
induction e; intros; subst; simpl in *.
  reflexivity.
  simpl. rewrite H. trivial. discriminate.
repeat rewrite zeval_app.
  rewrite IHe1. rewrite IHe2. trivial.
Qed.

Lemma zeval_upd_CONS: forall M n F p (HN: ~ In n F),
  zeval SVar (supd M (cond_svar n) p) (allPCs (n :: F)) = llub p (zeval SVar M (allPCs F)).
Proof.
simpl. intros.
(*rewrite zeval_CONS; simpl.*)
rewrite supd_Same.
rewrite <- zeval_allPCs_supd; trivial.
Qed.

Lemma zeval_app_pcVar_eq:
  forall (M M' : SVar -> L) (F : list nat) e,
  (forall z : SVar, z <> var_svar pcVar -> M z = M' z) ->
  zeval SVar M ((allPCs F)++(mkSVars e)) = zeval SVar M' ((allPCs F)++(mkSVars e)).
Proof.
intros.
repeat rewrite zeval_app.
rewrite (allPCs_pcVar_eq _ _ _ H). 
rewrite (mkSVars_pcVar_eq _ _ _ H).
trivial.
Qed.

(*Here is the difnition of the translation/synthesis iota*)
Fixpoint iota (C:Com) : (list nat) -> JCom SVar :=
match C with
  Skip => fun F => (JSkip SVar)
| Assign x e => fun F => 
       JComp SVar (JAssign _ x e) 
                  (JJoin _ (var_svar (progVar x)) nil ((allPCs F)++(mkSVars e)))
| Comp C1 C2 => fun F => JComp _ (iota C1 F) (iota C2 F)
| Ite e C1 C2 => (fun F => 
      match (fresh F) with exist n Pn => 
        let secregsPre := (allPCs F) ++ (mkSVars e) 
        in JComp _ (JJoin _ (cond_svar n) (e::nil) secregsPre  )
                   (JIte _ e (JComp _ (iota C1 (n::F)) (mkJoins n (modVarsL C2)))
                             (JComp _ (iota C2 (n::F)) (mkJoins n (modVarsL C1))))
      end)
| While e C => fun F => 
      match (fresh F) with exist n Pn => 
        let secregsPre := (allPCs F) ++ (mkSVars e) in
        let secregsInner := (allPCs (n::F))  ++ (mkSVars e)
        in (JComp _ (JJoin _ (cond_svar n) (e::nil) secregsPre)
                     (JComp _ (JWhile _ e (JComp _ (iota C (n::F))
                                            (JComp _ (mkJoins n (modVarsL C))
                                              (JJoin _ (cond_svar n) (e::nil) secregsInner))))
                              (mkJoins n (modVarsL C))))
        end
end.

Lemma Sem_iota_JErase: forall C s t (Hyp:Sem s C t) F,
  Sem s (JErase _ (iota C F)) t.
Proof.
intros C s t Hyp.
induction Hyp; intros; subst; simpl in *.
(*Skip*)  
  constructor.
 (*Assign*)
  eexists; eexists.
  econstructor.
  constructor; reflexivity.
  constructor; reflexivity.
(*Comp*)
  econstructor. apply IHHyp1. apply IHHyp2. 
(*IfTrue*)
  remember (fresh F) as FRn. destruct FRn as [n Hn].
  simpl.
  econstructor. econstructor. reflexivity.
  eapply S_True.
     apply H.
     econstructor. apply IHHyp. apply JErase_mkJoins_Sem.
(*IfFalse*)
  remember (fresh F) as FRn. destruct FRn as [n Hn].
  simpl.
  econstructor. econstructor. reflexivity.
  eapply S_False.
     apply H.
     econstructor. apply IHHyp. apply JErase_mkJoins_Sem.
(*WhileTrue*) 
  assert (IH := IHHyp2 F). clear IHHyp2.
  remember (fresh F) as FRn. destruct FRn as [n Hn].
  assert (IHyp1 := IHHyp1 (n::F)). clear IHHyp1.
  simpl in *.
  inversion IH; subst; clear IH.
  inversion H3; subst; clear H3. rename r0 into r.
  econstructor. constructor. reflexivity.
  inversion H5; subst; clear H5. rename r0 into u.
  econstructor.  
    eapply S_WTrue.
      apply H.
      econstructor. apply IHyp1.
      econstructor. apply JErase_mkJoins_Sem. 
                    constructor. reflexivity.
      apply H3.
    apply H6.
(*WhileFalse*) 
  remember (fresh F) as FRn. destruct FRn as [n Hn].
  simpl in *.
  econstructor. constructor. reflexivity.
  econstructor.  
    eapply S_WFalse.
      apply H.
      reflexivity. 
      apply JErase_mkJoins_Sem.
Qed. 

Lemma modVars_iota: forall C F, modVars (JErase _ (iota C F)) = modVars C. 
Proof. 
intros C. 
induction C; intros; simpl in *; trivial. 
(*Assign*)
  apply extensionality. intros x. apply Bool.orb_false_r.
(*Comp*)
  rewrite IHC1.  rewrite IHC2.  trivial.
(*Ite*) 
  remember (fresh F) as Fr. destruct Fr as [n Hn].
  simpl. rewrite IHC1.  rewrite IHC2. clear IHC1 IHC2. 
  repeat rewrite modVars_mkJoins. simpl.
  apply extensionality. intros x.
  repeat rewrite <- Bool.orb_assoc.
  rewrite Bool.orb_false_r.
  rewrite Bool.orb_false_l.
  trivial.
(*While*)
  remember (fresh F) as Fr. destruct Fr as [n Hn].
  simpl. rewrite IHC.
  apply extensionality. intros x.
  repeat rewrite modVars_mkJoins. simpl.
  repeat rewrite Bool.orb_false_r.
  trivial.
Qed.

(*First item of Lemma 5*)
Lemma iota_noMod: forall C F U 
  (HR: U = iota C F) 
  s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N)) x,
  modVars (JErase _ U) x = false ->
  M (var_svar (progVar x)) = N (var_svar (progVar x)).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem. trivial.
(*Assign*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H3; subst; clear H3. 
  repeat rewrite Bool.orb_false_r in H. apply pvar_eq_ElimF in H. 
  inversion H5; subst; clear H5.
  rewrite supd_Diff. trivial.
  intros Z. inversion Z; subst. apply H. trivial. 
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  apply Bool.orb_false_elim in H.
  destruct H as [MV1 MV2].
  rewrite <- (IHC2 _ _ (eq_refl _) _ _ _ _ _ _ H5 x MV2).
  apply (IHC1 _ _ (eq_refl _) _ _ _ _ _ _  H3 x MV1).
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [n FRn].
  simpl in H.
  apply Bool.orb_false_elim in H.
  destruct H as [MV1 MV2].
  apply Bool.orb_false_elim in MV2.
  destruct MV2 as [MV2a _].
  apply Bool.orb_false_elim in MV1.
  destruct MV1 as [MV1a _].
(*  rewrite modVars_iota in *.*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4; simpl in *.
(*  rewrite <- (mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H5). clear H5.*)
     (*True*) clear H5 IHC2.
        inversion H6; subst; clear H6. 
        destruct r0 as [[s K] Q].
        rewrite <- (mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H4). clear H4.
        rewrite <- (IHC1 _ _ (eq_refl _) _ _ _ _ _ _  H2 x). clear IHC1 H2.
        rewrite supd_Diff; try discriminate. trivial.
        apply MV1a.
        apply modVarsChar'. rewrite modVars_iota in MV2a. apply MV2a.
     (*False*) clear H5 IHC1.
        inversion H6; subst; clear H6. 
        destruct r0 as [[s K] Q].
        rewrite <- (mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H4). clear H4.
        rewrite <- (IHC2 _ _ (eq_refl _) _ _ _ _ _ _  H2 x). clear IHC2 H2.
        rewrite supd_Diff; try discriminate. trivial.
        apply MV2a.
        apply modVarsChar'. rewrite modVars_iota in MV1a. apply MV1a.
(*While*)
  remember (fresh F) as FRESH. destruct FRESH as [n FRn].
  simpl in H.
  apply Bool.orb_false_elim in H.
  destruct H as [MV1 _].
  apply Bool.orb_false_elim in MV1.
  destruct MV1 as [MV2 _].
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4. destruct r0 as [[s G] P].
  rewrite <- (mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H5). clear H5.
  assert (IHCC := IHC (n::F) _ (eq_refl _)). clear IHC.
  assert (IHW : forall (s : Store) (G : CTXT) (M : ZStore SVar) (t : Store) (D : CTXT)
         (N : ZStore SVar),
       JoinSem SVar SVar_eq (s, G, M) (JWhile SVar e
          (JComp SVar (iota C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JJoin SVar (cond_svar n) (e :: nil)
                   (cond_svar n :: allPCs F ++ mkSVars e))))) (t, D, N) ->
       forall x : PVar,
       modVars (JErase SVar (iota C (n :: F))) x = false ->
       M (var_svar (progVar x)) = N (var_svar (progVar x))).
    clear H2 HeqFRESH FRn. clear s G M t D N. clear r P E.
    remember (iota C (n :: F)) as R.
    intros s G M t D N HRS.
    remember (s, G, M) as ST.
    remember (t, D, N) as TT.
    remember (JWhile SVar e
           (JComp SVar R
              (JComp SVar (mkJoins n (modVarsL C))
                 (JJoin SVar (cond_svar n) (e :: nil)
                    (cond_svar n :: allPCs F ++ mkSVars e))))) as PP.
    generalize dependent N. generalize dependent D. generalize dependent t.
    generalize dependent M. generalize dependent G. generalize dependent s.
    induction HRS; intros; subst; try inversion HeqPP; subst; clear HeqPP.
    (*True*)
       simpl in *. clear IHHRS1. destruct r as [[r A] Q].
       inversion HRS1; subst; clear HRS1.  destruct r0 as [[u B] O].
       inversion H6; subst; clear H6. 
       inversion H8; subst; clear H8. 
       rewrite (IHCC _ _ _ _ _ _ H4 _ H0). clear H4 IHCC.
       assert (LP := IHHRS2 (eq_refl _)). clear IHHRS2.
       rewrite <- (LP _ _ _ (eq_refl _) _ _ _ (eq_refl _) _ H0). clear LP.
       rewrite supd_Diff; try discriminate.
       apply (mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H5).
         rewrite modVars_iota in H0. apply modVarsChar'. assumption.
       
    (*False*)
       inversion HeqTT; subst. trivial.
  clear IHCC.
    rewrite <- (IHW _ _ _ _ _ _ H2  _ MV2). clear IHW H2.
    rewrite supd_Diff; try discriminate. trivial.

  rewrite modVars_iota in MV2.
     intros Z. apply modVarsChar' in MV2. apply (MV2 Z).
Qed.

(*second item of Lemma 5*)
Lemma iota_ControlLabelsPreserved: forall C F U 
  (HR: U = iota C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N))
  n (IN: In n F),
  M (cond_svar n) = N (cond_svar n).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem. trivial.
(*Assign*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2. 
  inversion H4; subst; clear H4. 
  rewrite supd_Diff; try discriminate. trivial.
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  rewrite (IHC1 _ _ (eq_refl _) _ _ _ _ _ _ H2 _ IN).
  apply (IHC2 _ _ (eq_refl _) _ _ _ _ _ _ H4 _ IN).
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  (*True*)
     clear IHC2 H5.
     inversion H6; subst; clear H6. destruct r0 as [[s G] P].
     rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H4 n). clear H4.
       rewrite <- (IHC1 _ _ (eq_refl _) _ _ _ _ _ _ H2). clear IHC1 H2.
       rewrite supd_Diff. trivial.
       intros Z. inversion Z; subst. apply (FRm IN).
       right; apply  IN.
  (*False*)
     clear IHC1 H5.
     inversion H6; subst; clear H6. destruct r0 as [[s G] P].
     rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H4 n). clear H4.
       rewrite <- (IHC2 _ _ (eq_refl _) _ _ _ _ _ _ H2). clear IHC2 H2.
       rewrite supd_Diff. trivial.
       intros Z. inversion Z; subst. apply (FRm IN).
       right; apply  IN.
(*While*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4. destruct r0 as [[s G] P].
  rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H5 n).
  clear H5.
  assert (HC := IHC (m::F) _ (eq_refl _)). clear IHC.
  assert (IH: forall (s : Store) (G : CTXT) (M : ZStore SVar) (t : Store) (D : CTXT)
        (N : ZStore SVar),
      JoinSem _ SVar_eq (s, G, M) (JWhile SVar e
          (JComp SVar (iota C (m :: F))
             (JComp SVar (mkJoins m (modVarsL C))
                (JJoin SVar (cond_svar m) (e :: nil)
                   (cond_svar m :: allPCs F ++ mkSVars e))))) (t, D, N) ->
      forall n : nat, In n F -> M (cond_svar n) = N (cond_svar n)).

    clear H2 s G M r E P HeqFRESH t D N IN n.
    intros s G M t D N HRS.
    remember (s, G, M) as ST.
    remember (t, D, N) as TT.
    remember (JWhile SVar e
           (JComp SVar (iota C (m :: F))
              (JComp SVar (mkJoins m (modVarsL C))
                 (JJoin SVar (cond_svar m) (e :: nil)
                    (cond_svar m :: allPCs F ++ mkSVars e))))) as PP. 
    generalize dependent N. generalize dependent D. generalize dependent t.
    generalize dependent M. generalize dependent G. generalize dependent s.
    induction HRS; intros; subst; try inversion HeqPP; subst; clear HeqPP.
    (*True*)
       simpl in *. clear IHHRS1. destruct r as [[r A] Q].
       inversion HRS1; subst; clear HRS1.
       inversion H6; subst; clear H6.
       inversion H8; subst; clear H8.
       rewrite <- (IHHRS2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _) _ H0). clear IHHRS2.
       rewrite supd_Diff.
         destruct r0 as [[u B] Q].
         rewrite (HC _ _ _ _ _ _ H4).
         rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H5 n). trivial.
         right; assumption.
       intros Z. inversion Z; subst. apply (FRm H0).
    (*False*)
       inversion HeqTT; subst. trivial.
  clear HC.
    rewrite <- (IH _ _ _ _ _ _ H2). clear IH H2.
    rewrite supd_Diff. trivial.
     intros Z. inversion Z; subst. apply (FRm IN).
    assumption. 
Qed.

(*third item of Lemma 5*)
Lemma iota_Mod: forall C F U
  (Hiota: U = iota C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N)) x
  (MVx: modVars (JErase _ U) x = true) n
  (IN: In n F),
  ssqsub (N (cond_svar n)) (N (var_svar (progVar x))).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion MVx.
(*Assign*)
  rewrite Bool.orb_false_r in MVx. apply pvar_eq_ElimT in MVx. subst.
  inversion HRSem; subst; clear HRSem.
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  rewrite supd_Same.
  rewrite supd_Diff; try discriminate.
    apply lubright. 
    apply zeval_IN_svar. apply in_or_app. left. apply allPCs_In. apply IN.
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  apply Bool.orb_true_iff in MVx.
  remember (modVars (JErase _ (iota C2 F)) x) as b2; destruct b2; subst; simpl in *; apply eq_sym in Heqb2. 
  (*modVars SVar R2 x = true*)
     eapply (IHC2 F _ (eq_refl _)). apply H4. apply Heqb2. apply IN. 
  (*modVars SVar R2 x = false*)
     rewrite <- (iota_noMod C2 F _ (eq_refl _) _ _ _ _ _ _ H4 _ Heqb2).
     rewrite <- (iota_ControlLabelsPreserved C2 F _ (eq_refl _) _ _ _ _ _ _ H4 _ IN).
     clear H4 IHC2.
     destruct MVx as [MV | MV]; try inversion MV.
     eapply (IHC1 F _ (eq_refl _)). apply H2. apply MV. apply IN.
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  simpl in MVx.
  repeat rewrite modVars_mkJoins in MVx. 
  repeat rewrite Bool.orb_false_r in MVx.
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4; simpl in *.
  (*True*) clear IHC2.
     inversion H6; subst; clear H6. destruct r0 as [[s G] P].
       destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.

       remember (in_dec pvar_eq x (modVarsL C2)) as b; destruct b; simpl.
          remember (in_dec pvar_eq x (modVarsL C1)) as bb; destruct bb; simpl.
          eapply leq_trans. eapply (IHC1 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2 x). simpl. 
                               rewrite modVars_iota. apply modVarsChar. assumption.
                               right; assumption.
                            apply lub_leqright.


       assert (Z:= iota_ControlLabelsPreserved C1 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2).
        rewrite <- (Z n). rewrite supd_Diff. Focus 2. intros ZZ. inversion ZZ. subst. apply (FRm IN).
        assert (Y : In m (m::F)). left. trivial.
           assert (ZZ := Z _ Y). clear Z Y. rewrite supd_Same in ZZ.
           eapply (@leq_trans L LatL _ (zeval SVar M (allPCs F ++ mkSVars e))).
             apply zeval_IN_svar. apply in_or_app. left. apply allPCs_In. assumption.
           eapply (@leq_trans L LatL _ (llub (lab_eval E (e :: nil)) (zeval SVar M (allPCs F ++ mkSVars e)))).
              apply lub_leqright.
           rewrite <- ZZ. apply lubleft. apply leq_refl.
        right. assumption.

        remember (in_dec pvar_eq x (modVarsL C1)) as bb; destruct bb; simpl.
          eapply leq_trans. eapply (IHC1 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2 x). simpl. 
                               rewrite modVars_iota. apply modVarsChar. assumption.
                               right; assumption.
                            apply leq_refl.
         remember (modVars C1 x) as cc; destruct cc; apply eq_sym in Heqcc. rewrite modVarsChar in Heqcc. exfalso. apply (n1 Heqcc).          
         remember (modVars C2 x) as dd; destruct dd; apply eq_sym in Heqdd. rewrite modVarsChar in Heqdd. exfalso. apply (n0 Heqdd).
         repeat rewrite modVars_iota in MVx. rewrite Heqcc in MVx. rewrite Heqdd in MVx. simpl in MVx. inversion MVx.

   (*False*) clear IHC1.
     inversion H6; subst; clear H6. destruct r0 as [[s G] P].
       destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.

       remember (in_dec pvar_eq x (modVarsL C1)) as b; destruct b; simpl.
          remember (in_dec pvar_eq x (modVarsL C2)) as bb; destruct bb; simpl.
          eapply leq_trans. eapply (IHC2 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2 x). simpl. 
                               rewrite modVars_iota. apply modVarsChar. assumption.
                               right; assumption.
                            apply lub_leqright.


       assert (Z:= iota_ControlLabelsPreserved C2 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2).
        rewrite <- (Z n). rewrite supd_Diff. Focus 2. intros ZZ. inversion ZZ. subst. apply (FRm IN).
        assert (Y : In m (m::F)). left. trivial.
           assert (ZZ := Z _ Y). clear Z Y. rewrite supd_Same in ZZ.
           eapply (@leq_trans L LatL _ (zeval SVar M (allPCs F ++ mkSVars e))).
             apply zeval_IN_svar. apply in_or_app. left. apply allPCs_In. assumption.
           eapply (@leq_trans L LatL _ (llub (lab_eval E (e :: nil)) (zeval SVar M (allPCs F ++ mkSVars e)))).
              apply lub_leqright.
           rewrite ZZ. apply lubleft. apply leq_refl.
        right. assumption.

        remember (in_dec pvar_eq x (modVarsL C2)) as bb; destruct bb; simpl.
          eapply leq_trans. eapply (IHC2 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2 x). simpl. 
                               rewrite modVars_iota. apply modVarsChar. assumption.
                               right; assumption.
                            apply leq_refl.
         remember (modVars C1 x) as cc; destruct cc; apply eq_sym in Heqcc. rewrite modVarsChar in Heqcc. exfalso. apply (n0 Heqcc).          
         remember (modVars C2 x) as dd; destruct dd; apply eq_sym in Heqdd. rewrite modVarsChar in Heqdd. exfalso. apply (n1 Heqdd).
          repeat rewrite modVars_iota in MVx. rewrite Heqcc in MVx. rewrite Heqdd in MVx. simpl in MVx. inversion MVx.
(*While*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  simpl in MVx. repeat rewrite modVars_iota in MVx.
  repeat rewrite modVars_mkJoins in MVx. 
  repeat rewrite Bool.orb_false_l in MVx. 
  repeat rewrite Bool.orb_false_r in MVx. 
  assert (IH := IHC (m::F) _ (eq_refl _)). clear IHC.
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4. destruct r0 as [[s G] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
  remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b; subst; simpl in *.
    Focus 2. apply modVarsChar in MVx. exfalso. apply (n0 MVx).
  apply lubleft.
  clear IH.
  assert (Z : forall ST TT,
         JoinSem _ SVar_eq ST (JWhile SVar e
          (JComp SVar (iota C (m :: F))
             (JComp SVar (mkJoins m (modVarsL C))
                (JJoin SVar (cond_svar m) (e :: nil)
                   (cond_svar m :: allPCs F ++ mkSVars e))))) TT ->
           ((snd ST) (cond_svar n) = (snd TT) (cond_svar n) /\
            ssqsub ((snd ST) (cond_svar m)) ((snd TT) (cond_svar m)))).
    clear H2. clear Heqb i r E P HeqFRESH MVx x s G M.
    intros ST TT HRS. remember (JWhile SVar e
           (JComp SVar (iota C (m :: F))
              (JComp SVar (mkJoins m (modVarsL C))
                 (JJoin SVar (cond_svar m) (e :: nil)
                    (cond_svar m :: allPCs F ++ mkSVars e))))) as PP.
    induction HRS; subst; try inversion HeqPP; subst; clear HeqPP; simpl in *.
     (*True*) clear IHHRS1.
              destruct (IHHRS2 (eq_refl _)) as [Z1 Z2]. clear IHHRS2 HRS2.
              destruct s as [[s G] M]. destruct r as [[r D] N].
              inversion HRS1; subst; clear HRS1. destruct r0 as [[u E] P]. simpl in *.
              inversion H5; subst; clear H5.
              inversion H7; subst; clear H7.
              simpl in *. rewrite supd_Same in Z2.
              rewrite supd_Diff in Z1.
                assert (Z:= iota_ControlLabelsPreserved C (m::F) _ (eq_refl _) _ _ _ _ _ _ H3). clear H3.
                assert (ZZ:= mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H4).
                rewrite (Z n). rewrite <- Z1. split. apply ZZ.
                   rewrite (Z m). eapply leq_snart. apply Z2. apply lubright. rewrite (ZZ m). apply lub_leqleft. left. trivial.
                right; assumption.
              intros Z. inversion Z; subst. apply (FRm IN).
      (*False*) split. reflexivity. apply leq_refl.
   destruct (Z _ _ H2) as [Z1 Z2]; clear Z H2. simpl in *.
   rewrite supd_Same in Z2. 
   rewrite supd_Diff in Z1. rewrite <- Z1. clear Z1.
          eapply leq_snart. apply Z2. apply lubright. apply zeval_IN_svar. apply in_or_app. left. apply allPCs_In. trivial.        
   intros Z. inversion Z; subst. apply (FRm IN).
Qed.

Lemma iota_Mod': forall C F U
  (Hiota: U = iota C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N)) x
  (MVx: modVars (JErase _ U) x = true) n
  (IN: In n F),
  ssqsub (M (cond_svar n)) (N (var_svar (progVar x))).
Proof.
intros. subst.
rewrite (iota_ControlLabelsPreserved C F _ (eq_refl _) _ _ _ _ _ _ HRSem).
  eapply (iota_Mod C F _ (eq_refl _)). apply HRSem. assumption. assumption.
assumption.
Qed.

Lemma JoinSem_iota_allPCs: forall C G M H N F s t
(Hyp: JoinSem _ SVar_eq (s,G,M) (iota C F) (t,H,N)),
zeval SVar M (allPCs F) = zeval SVar N (allPCs F).
Proof.
intros.
assert (Z:= iota_ControlLabelsPreserved C F _ (eq_refl _) _ _ _ _ _ _ Hyp). clear Hyp.
induction F; simpl; intros; subst.
  trivial.
  rewrite IHF. rewrite (Z a). trivial. left. trivial.
  intros. apply Z. right. trivial.
Qed.

Lemma iota_JoinSem_Sem: forall C F s G M t D N
(Hyp: JoinSem _ SVar_eq (s,G,M) (iota C F) (t,D,N)), 
  Sem s C t. 
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)  
  inversion Hyp; subst; simpl; clear Hyp.
  constructor.  trivial. 
 (*Assign*)
  inversion Hyp; subst; simpl; clear Hyp.
  inversion H2; subst; simpl; clear H2.
  inversion H4; subst; simpl; clear H4.
  constructor.  trivial. 
(*Comp*)
  inversion Hyp; subst; simpl; clear Hyp. 
  destruct r as [[r K] P]. 
  assert (IH1 := IHC1 _ _ _ _ _ _ _ H2). 
  assert (IH2 := IHC2 _ _ _ _ _ _ _ H4). 
  econstructor.  apply IH1.  assumption. 
(*If*)
  remember (fresh F) as FRn. destruct FRn as [n Hn].
  inversion Hyp; subst; simpl; clear Hyp.
  inversion H2; subst; simpl; clear H2.
  inversion H4; subst; simpl; clear H4; simpl in *.
  (*True*) clear IHC2.
     inversion H6; subst; simpl; clear H6. destruct r as [[r K] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]; clear H4.
     subst.
     assert (IH1 := IHC1 _ _ _ _ _ _ _ H2).
     apply S_True; assumption.
  (*False*) clear IHC1.
     inversion H6; subst; simpl; clear H6. destruct r as [[r K] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]; clear H4.
     subst.
     assert (IH1 := IHC2 _ _ _ _ _ _ _ H2).
     apply S_False; assumption.
(*While*)
  remember (fresh F) as FRn. destruct FRn as [n Hn].
  inversion Hyp; subst; simpl; clear Hyp.
  inversion H2; subst; simpl; clear H2.
  inversion H4; subst; simpl; clear H4. destruct r as [[r K] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]; clear H5.
  subst.
  remember ((JWhile SVar e
          (JComp SVar (iota C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JJoin SVar (cond_svar n) (e :: nil)
                   (cond_svar n :: allPCs F ++ mkSVars e)))))) as LOOP.
  remember (s, G,
       zupd SVar SVar_eq M (cond_svar n)
         (llub (lab_eval G (e :: nil)) (zeval SVar M (allPCs F ++ mkSVars e)))) as ST.
  remember  (r, K, P) as TT.
  assert (IH: Sem (fst (fst ST)) (While e C) (fst (fst TT))).
    clear HeqST HeqTT M G s P K r.
    induction H2; intros; try inversion HeqLOOP; subst; clear HeqLOOP; simpl in *.
    (*True*) clear H2_0 IHJoinSem1.
             destruct r as [[r K] P]. destruct s as [[s G] M]. destruct t as [[t D] N]; simpl in *.
             inversion H2_; subst; clear H2_.
             inversion H5; subst; clear H5.
             inversion H7; subst; clear H7. destruct r0 as [[u A] Q].
             destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]; clear H4.
             subst.
             assert (IH := IHC _ _ _ _ _ _ _ H3).
             eapply S_WTrue; try assumption. apply IH. apply (IHJoinSem2 (eq_refl _)).
    (*False*) destruct t as [[t D] N]; simpl in *.
             eapply S_WFalse; trivial.
  subst; simpl in *. apply IH.
Qed.

Lemma iota_Sem_JoinSem: forall C s t (Hyp: Sem s C t) F G M,
  exists D, exists N, JoinSem _ SVar_eq (s,G,M) (iota C F) (t,D,N).
Proof.
intros.
assert (X:= Sem_iota_JErase C s t Hyp F).
assert (Z:= JErase_Lemma_Inv _ SVar_eq (iota C F) _ _ X G M).
apply Z.
Qed.  

Lemma iota_Sem: forall C F U
  (Hiota: U = iota C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N)) s' t'
  (HSem: Sem s' C t') x,
  (forall y, ssqsub (llub (G y) (M (var_svar (progVar y))))
                    (llub (D x) (N (var_svar (progVar x)))) 
             -> s y = s' y) ->
  t x = t' x.
Proof.
intros C. 
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem.
  inversion HSem; subst; clear HSem.
  apply H. apply leq_refl.
(*Assign*) 
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5.
  inversion HSem; subst; clear HSem.
  apply M_upd_split_at; intros; subst.
     rewrite M_upd_Same. repeat rewrite M_upd_Same in H.
     eapply (eq_below_evalE _ (fun z => llub (G z) (P (var_svar (progVar z)))))
          with (kappa:=(GEval (fun z : PVar => llub (G z) (P (var_svar (progVar z)))) e)).
       unfold eq_below. intros z Hz. apply H.
         eapply leq_trans. apply Hz. clear Hz H.
         rewrite GEval_sep.
         apply lub_leq. apply lub_leqleft.
         apply lubright. rewrite supd_Same. rewrite zeval_app. apply lubright. rewrite zeval_GEval. apply lub_leqright.
       apply leq_refl.
   rewrite M_upd_Diff; try assumption.
     rewrite M_upd_Diff in H; try assumption.
     rewrite supd_Diff in H. apply H; clear H. apply leq_refl.
      intros Z; inversion Z; subst. apply H0; trivial.
(*Comp*)
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion HSem; subst; clear HSem.
  remember r0 as r'. clear Heqr'.
  eapply (IHC2 F _ (eq_refl _)); clear IHC2. apply H5. apply H7.
  intros y Hy.
  eapply (IHC1 F _ (eq_refl _)); clear IHC1. apply H3. apply H4.
  intros z Hz.
  apply H; clear H.
  eapply leq_trans. apply Hz. apply Hy.
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [n FRn].
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5; simpl in *.
  (*eval s e = True*) clear IHC2.
     inversion H7; subst; clear H7. destruct r0 as [[s G] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]; clear H5.
     subst.
     inversion HSem; subst; clear HSem.
    (*eval s' e = True*)
        eapply (IHC1 (n::F) _ (eq_refl _)). apply H3. apply H8.
        intros y Hy. apply H; clear H.
        rewrite supd_Diff in Hy; try discriminate.
        eapply leq_trans. apply Hy. clear Hy.
        apply lub_leq. apply lub_leqleft.
        apply lubright. remember (in_dec pvar_eq x (modVarsL C2)) as b; destruct b; subst; simpl in *.
           apply lub_leqright.
           apply leq_refl.
     (*eval s' e = False*)
        clear IHC1.
        remember (in_dec pvar_eq x (modVarsL C1)) as b; destruct b; subst; simpl in *.
        (*In x modVars C1*) 
             assert (MVx :modVars (JErase _ (iota C1 (n :: F))) x = true).
               rewrite modVars_iota. apply modVarsChar. assumption.
             assert (Z:= iota_Mod' C1 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 _ MVx n).
             rewrite supd_Same in Z.
             assert (EV: eval r e = eval s' e).
             eapply (eq_below_evalE _ (fun z => llub (E z) (M (var_svar (progVar z)))))
                 with (kappa:=(GEval (fun z : PVar => llub (E z) (M (var_svar (progVar z)))) e)).
                 intros z Hz. apply H. clear H. 
                 eapply leq_trans. apply Hz. clear Hz.
                 apply lubright.
                 apply @leq_snart with (y:= P (var_svar (progVar x))).
                   remember (in_dec pvar_eq x (modVarsL C2)) as bb; destruct bb; subst; simpl in *.
                      apply lub_leqright.
                      apply leq_refl.
                 eapply leq_snart. apply Z. left. trivial. 
                   clear Z. rewrite GEval_sep. 
                   apply lub_leq. apply lubleft. rewrite lab_eval_GEval. apply leq_refl.
                   apply lubright. rewrite zeval_app. apply lubright. rewrite zeval_GEval. apply leq_refl.
               apply leq_refl.
             rewrite EV in H6. rewrite H6 in H7. inversion H7.
         (*~ In x (MV C1*)
           assert (MVx :modVars (JErase _ (iota C1 (n :: F))) x = false).
             rewrite modVars_iota. apply modVarsChar'. assumption.
           destruct (JSem1 _ _ _ _ _ _ _ _ _ H3 _ MVx) as [Z1 Z2]. 
           rewrite <- Z1 in *. clear Z1.
           rewrite <- Z2 in *. clear Z2. 
           remember (modVars C2 x) as bb; destruct bb; subst; simpl in *; apply eq_sym in Heqbb.
             remember (in_dec pvar_eq x (modVarsL C2)) as bbb; destruct bbb; subst; simpl in *; apply eq_sym in Heqbbb.
               assert (EV: eval r e = eval s' e).                      
                 eapply (eq_below_evalE _ (fun z => llub (E z) (M (var_svar (progVar z)))))
                     with (kappa:=(GEval (fun z : PVar => llub (E z) (M (var_svar (progVar z)))) e)).
                   intros z Hz. apply H. clear H. 
                   eapply leq_trans. apply Hz. clear Hz.
                   assert (Z:= iota_ControlLabelsPreserved C1 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 n). clear H3.
                   rewrite supd_Same in Z. 
                   apply lubright. apply lubleft.
                   rewrite <- Z. 
                     rewrite GEval_sep. apply lub_leq. apply lubleft. rewrite lab_eval_GEval. apply leq_refl.
                     apply lubright. rewrite zeval_app. rewrite <- zeval_GEval. apply lub_leqright.
                   left; trivial. 
                 apply leq_refl.
               rewrite EV in H6. rewrite H6 in H7. inversion H7.
             exfalso. apply n1. apply modVarsChar. apply Heqbb.
           rewrite <- (Sem_nomod _ _ _ H8 _ Heqbb). 
             apply H. clear H H8.  
             remember (in_dec pvar_eq x (modVarsL C2)) as bbb; destruct bbb; subst; simpl in *; apply eq_sym in Heqbbb.
                exfalso. clear Heqbbb. apply modVarsChar in i. rewrite i in Heqbb. inversion Heqbb. 
             assert (Z:= iota_noMod C1 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 _ MVx).
                    rewrite supd_Diff in Z; try discriminate. rewrite Z. apply leq_refl. 
  (*eval s e = False*) clear IHC1.
     inversion H7; subst; clear H7. destruct r0 as [[s G] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]; clear H5.
     subst.
     inversion HSem; subst; clear HSem.
     (*eval s' e = true*) clear IHC2.
        remember (in_dec pvar_eq x (modVarsL C2)) as b; destruct b; subst; simpl in *.
        (*In x modVars C2*) 
             assert (MVx : modVars (JErase _ (iota C2 (n :: F))) x = true).
               rewrite modVars_iota. apply modVarsChar. assumption.
             assert (Z:= iota_Mod' C2 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 _ MVx n).
             rewrite supd_Same in Z.
             assert (EV: eval r e = eval s' e).
             eapply (eq_below_evalE _ (fun z => llub (E z) (M (var_svar (progVar z)))))
                 with (kappa:=(GEval (fun z : PVar => llub (E z) (M (var_svar (progVar z)))) e)).
                 intros z Hz. apply H. clear H. 
                 eapply leq_trans. apply Hz. clear Hz.
                 apply lubright.
                 apply @leq_snart with (y:= P (var_svar (progVar x))).
                   remember (in_dec pvar_eq x (modVarsL C1)) as bb; destruct bb; subst; simpl in *.
                      apply lub_leqright.
                      apply leq_refl.
                 eapply leq_snart. apply Z. left. trivial. 
                   clear Z. rewrite GEval_sep. 
                   apply lub_leq. apply lubleft. rewrite lab_eval_GEval. apply leq_refl.
                   apply lubright. rewrite zeval_app. apply lubright. rewrite zeval_GEval. apply leq_refl.
               apply leq_refl.
             rewrite EV in H6. rewrite H6 in H7. inversion H7.
         (*~ In x (MV C2*)
           assert (MVx : modVars (JErase _ (iota C2 (n :: F))) x = false).
             rewrite modVars_iota. apply modVarsChar'. assumption.
           destruct (JSem1 _ _ _ _ _ _ _ _ _ H3 _ MVx) as [Z1 Z2]. 
           rewrite <- Z1 in *. clear Z1.
           rewrite <- Z2 in *. clear Z2. 
           remember (modVars C1 x) as bb; destruct bb; subst; simpl in *; apply eq_sym in Heqbb.
             remember (in_dec pvar_eq x (modVarsL C1)) as bbb; destruct bbb; subst; simpl in *; apply eq_sym in Heqbbb.
               assert (EV: eval r e = eval s' e).                      
                 eapply (eq_below_evalE _ (fun z => llub (E z) (M (var_svar (progVar z)))))
                     with (kappa:=(GEval (fun z : PVar => llub (E z) (M (var_svar (progVar z)))) e)).
                   intros z Hz. apply H. clear H. 
                   eapply leq_trans. apply Hz. clear Hz.
                   assert (Z:= iota_ControlLabelsPreserved C2 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 n). clear H3.
                   rewrite supd_Same in Z. 
                   apply lubright. apply lubleft.
                   rewrite <- Z. 
                     rewrite GEval_sep. apply lub_leq. apply lubleft. rewrite lab_eval_GEval. apply leq_refl.
                     apply lubright. rewrite zeval_app. rewrite <- zeval_GEval. apply lub_leqright.
                   left; trivial. 
                 apply leq_refl.
               rewrite EV in H6. rewrite H6 in H7. inversion H7.
             exfalso. apply n1. apply modVarsChar. apply Heqbb.
           rewrite <- (Sem_nomod _ _ _ H8 _ Heqbb). 
             apply H. clear H H8.  
             remember (in_dec pvar_eq x (modVarsL C1)) as bbb; destruct bbb; subst; simpl in *; apply eq_sym in Heqbbb.
                exfalso. clear Heqbbb. apply modVarsChar in i. rewrite i in Heqbb. inversion Heqbb. 
             assert (Z:= iota_noMod C2 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 _ MVx).
                    rewrite supd_Diff in Z; try discriminate. rewrite Z. apply leq_refl.
     (*eval s' e = false*) 
        eapply (IHC2 (n::F) _ (eq_refl _)). apply H3. apply H8.
        intros y Hy. apply H; clear H.
        rewrite supd_Diff in Hy; try discriminate.
        eapply leq_trans. apply Hy. clear Hy.
        apply lub_leq. apply lub_leqleft.
        apply lubright. remember (in_dec pvar_eq x (modVarsL C1)) as b; destruct b; subst; simpl in *.
           apply lub_leqright.
           apply leq_refl.
(*While*)
  remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b; subst; simpl in *.
Focus 2. 
(*~In x (modVarsL C)*)
  clear IHC.
  assert (Z:= iota_noMod (While e C) F _ (eq_refl _)). simpl in Z.
  assert (ZZ:= Z _ _ _ _ _ _ HRSem x). clear Z.
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion H3; subst; clear H3. rename r into s. rename E into G.
  inversion H5; subst; clear H5. destruct r as [[r E] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]; clear H6. 
  apply eq_sym in Z1; apply eq_sym in Z2; subst.
  clear Heqb. 
  assert (MV: modVars C x = false). apply modVarsChar'. apply n.
  simpl in ZZ. rewrite modVars_mkJoins in ZZ. rewrite Bool.orb_false_l in ZZ. repeat rewrite Bool.orb_false_r in ZZ.
         rewrite modVars_iota in ZZ.
  destruct (JSem1 _ _ _ _ _ _ _ _ _ H3 x) as [X1 X2].
    simpl. rewrite modVars_mkJoins. rewrite modVars_iota. rewrite MV.  reflexivity.
  rewrite <- X1. clear X1.
  rewrite <- (Sem_nomod _ _ _ HSem).
     clear H3. apply H. rewrite X2. 
     remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b; subst; simpl.
       exfalso. apply (n i).
       rewrite ZZ. apply leq_refl.
      apply MV.
  apply MV.
(*In x (modVarsL C)*) 
  remember (fresh F) as FRESH. destruct FRESH as [n FRn].
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion H3; subst; clear H3. rename r into s. rename E into G.
  inversion H5; subst; clear H5. destruct r as [[r E] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]; clear H6.
  apply eq_sym in Z1; apply eq_sym in Z2; subst.
  repeat rewrite <- Heqb in *.
  assert (IH := IHC (n::F) _ (eq_refl _)); clear IHC.

  assert (SUB: forall s G M t E P 
              (HR: JoinSem _ SVar_eq (s, G, M)
                       (JWhile SVar e
          (JComp SVar (iota C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JJoin SVar (cond_svar n) (e :: nil)
                   (cond_svar n :: allPCs F ++ mkSVars e))))) 
                       (t, E, P)),
                   ssqsub (M (cond_svar n)) (P (cond_svar n))).
  (*Proof of SUB*)
    clear H3 H HSem s G M t P D s' t'. intros.
      remember (s, G, M) as ST.
      remember (t, E, P) as TT.
      remember (JWhile SVar e
          (JComp SVar (iota C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JJoin SVar (cond_svar n) (e :: nil)
                   (cond_svar n :: allPCs F ++ mkSVars e))))) as PP.
      generalize dependent P. generalize dependent E. generalize dependent t.
      generalize dependent M. generalize dependent G. generalize dependent s.
      induction HR; intros; subst; try inversion HeqPP; subst; clear HeqPP.
      (*True*) clear IHHR1 HR2 IH; simpl in *.
               inversion HR1; subst; clear HR1.
               inversion H5; subst; clear H5.
               inversion H7; subst; clear H7.
               assert (IHH := IHHR2 (eq_refl _) _ _ _  (eq_refl _) _ _ _ (eq_refl _)). clear IHHR2.
               eapply leq_snart. apply IHH.
               clear IHH.
               rewrite supd_Same.
               destruct r0 as [[r B] Q].
               apply lubright.
               eapply leq_snart. apply zeval_IN. left. reflexivity.
               simpl. rewrite (iota_ControlLabelsPreserved C (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 n).
                      rewrite (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H4).
                      apply leq_refl.
               left; trivial.
       (*False*) inversion HeqTT; subst; clear HeqTT.
                 apply leq_refl.
  (*Use of SUB*)
  remember (JWhile SVar e
           (JComp SVar (iota C (n :: F))
              (JComp SVar (mkJoins n (modVarsL C))
                 (JJoin SVar (cond_svar n) (e :: nil)
                    (cond_svar n :: allPCs F ++ mkSVars e))))) as PP.
  cut (forall (s : Store) (GG : CTXT) (MM : ZStore SVar) (t : Store) (D : CTXT)
       (N : ZStore SVar)
     (HR: JoinSem _ SVar_eq (s, GG, MM) PP (t, D, N))
     s' t' 
     (HS: Sem s' (While e C) t')
     (HM: ssqsub (llub (lab_eval GG (e :: nil)) (zeval SVar MM (allPCs F ++ mkSVars e)))
                 (MM (cond_svar n))),
     (forall y : PVar,
      ssqsub (llub (GG y) (MM (var_svar (progVar y))))
        (llub (D x) (llub (N (cond_svar n)) (N (var_svar (progVar x))))) -> s y = s' y) -> t x = t' x).
  (*Use OF CUT*)
     intros CUT.
     eapply CUT; clear CUT.
       apply H3.
       apply HSem. 
       rewrite supd_Same. apply lub_leq. apply lub_leqleft.
          apply lubright. repeat rewrite zeval_all_app. apply leq_refl. intros Z. apply FRn. apply Z.
     intros y Hy. rewrite supd_Diff in Hy; try discriminate.
         apply H. apply Hy.
  (*Proof OF CUT*)
   clear H3 HSem H s s' t t' D.
   intros ss GG MM tt DD NN HR.
   remember (ss, GG,MM) as ST. remember (tt, DD,NN) as TT.
   generalize dependent NN. generalize dependent DD. generalize dependent tt.
   generalize dependent MM. generalize dependent GG. generalize dependent ss.
   induction HR; intros; subst; try inversion HeqPP; subst; clear HeqPP.
   (*OUTER True*) clear IHHR1; simpl in *. destruct r as [[r B] Q].
       assert (IHH := IHHR2 (eq_refl _)); clear IHHR2.
       assert (SUB_IH := SUB _ _ _ _ _ _ HR2). clear HR2.
       inversion HR1; subst; clear HR1. destruct r0 as [[u A] K].
       inversion H6; subst; clear H6.
       inversion H8; subst; clear H8.

       assert (EV: eval ss e = eval s' e). clear IHH SUB IH HS.
         eapply eq_below_evalE with (G:= (fun z => llub (GG z) (MM (var_svar (progVar z)))))
              (kappa:=(GEval (fun z : PVar => llub (GG z) (MM (var_svar (progVar z)))) e)).
           intros y Hy. apply H0; clear H0.
             apply (leq_trans Hy); clear Hy.
             apply lubright. apply lubleft. 
              eapply leq_snart. apply SUB_IH.
              clear SUB_IH.
              rewrite supd_Same.
              rewrite (iota_ControlLabelsPreserved C (n::F) _ (eq_refl _) _ _ _ _ _ _ H4 n) in HM;  clear H4.
                apply lubright. eapply leq_snart. apply zeval_IN. left. reflexivity.
                simpl. rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H5 n).
                eapply leq_snart. apply HM.
                rewrite GEval_sep.  
                apply lub_leq. rewrite lab_eval_GEval. apply lub_leqleft.
                   apply lubright. rewrite zeval_app. rewrite zeval_GEval. apply lub_leqright.
             left; trivial.
           apply leq_refl.
       inversion HS; subst; clear HS SUB_IH HM.
       (*INNER True*) remember r as u'. clear Hequ'.
             apply (IHH SUB _ _ _ (eq_refl _) _ _ _ (eq_refl _) _ _ H9); clear IHH SUB H9.
                rewrite supd_Same.
                apply lub_leq. apply lub_leqleft.
                apply lubright. repeat rewrite zeval_all_app. apply zeval_SUBLIST. intros z. right. assumption.
                intros Z. apply FRn. apply Z.
             intros y Hy. 
               rewrite supd_Diff in Hy; try discriminate.
               destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. 
               simpl in *. 
               eapply IH; clear IH. apply H4. apply H7.
                 intros z Hz. apply H0; clear H0.
                   apply (leq_trans Hz); clear Hz. 
                   eapply leq_snart. apply Hy.
                    apply lub_leq. apply lub_leqleft.
                    apply lubright. remember (in_dec pvar_eq y (modVarsL C)) as bb; destruct bb; subst; simpl.
                        apply lub_leqright. apply leq_refl.
       (*INNER False*) clear IHH SUB IH H0 H4.
              rewrite EV in H. rewrite H in H6. inversion H6.
   (*OUTER False*) clear SUB IH.
       apply eq_sym in HeqTT. inversion HeqTT; subst; clear HeqTT.
       simpl in *.
       assert (EV: eval ss e = eval s' e).
         apply eq_below_evalE with (G:= (fun z => llub (GG z) (MM (var_svar (progVar z)))))
                      (kappa:=(GEval (fun z : PVar => llub (GG z) (MM (var_svar (progVar z)))) e)).
           intros y Hy. apply H1; clear H1.
             apply (leq_trans Hy); clear Hy.
             apply lubright. apply lubleft.
             eapply leq_snart. apply HM.
             rewrite GEval_sep.
             apply lub_leq. rewrite lab_eval_GEval. apply lub_leqleft.
                   apply lubright. rewrite zeval_app. rewrite zeval_GEval. apply lub_leqright.
           apply leq_refl.
       inversion HS; subst; clear HS HM.
       (*INNER True*) rewrite EV in H. rewrite H in H3. inversion H3.
       (*INNER False*) apply H1; clear H1. 
         apply lub_leq. apply lub_leqleft.
         apply lubright. apply lub_leqright.
Qed.

Definition MERGE (G : CTXT) (M : ZStore SVar) : CTXT :=
fun x => llub (G x) (M (var_svar (progVar x))).

Lemma MERGE_supd: forall G M n p, MERGE G (supd M (cond_svar n) p) = MERGE G M.
Proof.
intros.
apply extensionality.
intros z. unfold MERGE. rewrite supd_Diff. trivial. discriminate. 
Qed. 

Lemma MERGE_L: forall G H, SQSUB G (MERGE G H).
Proof.
intros G H z.
apply lub_leqleft. 
Qed. 

Lemma MERGE_R: forall G H, SQSUB (fun x : PVar => H (var_svar (progVar x))) (MERGE G H).
Proof.
intros G H z.
apply lub_leqright. 
Qed.

Lemma MERGE_monoleft: forall M M' G, SQSUB M M' -> SQSUB (MERGE M G) (MERGE M' G).
Proof.
intros.
intros z.
apply leq_monoleft. apply H.
Qed.

Lemma MERGE_monoright: forall M G G', (forall z, ssqsub (G z) (G' z)) -> SQSUB (MERGE M G) (MERGE M G').
Proof.
intros.
intros z.
apply leq_monoright. apply H.
Qed.

Lemma MERGE_pcupd:
  forall (G : CTXT) (M : SVar -> L) (p : L),
  MERGE G (supd M (var_svar pcVar) p) = MERGE G M.
Proof.
intros.
apply extensionality. intros z.
unfold MERGE; simpl. 
rewrite supd_Diff; congruence.
Qed.

Lemma zeval_lub_MERGE: forall G M F E,
      llub (GEval G E) (zeval SVar M (allPCs F ++ mkSVars E)) =
      llub (GEval (MERGE G M) E) (zeval SVar M (allPCs F)).
Proof.
intros; simpl.
rewrite zeval_app.
unfold MERGE; simpl.
rewrite GEval_sep; simpl.
rewrite zeval_GEval.
rewrite (llub_com (zeval _ M (allPCs F))).
apply llub_assoc. 
Qed.

Lemma Theorem4: forall C F s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (iota C F) (t,D,N)) s' t'
  (HSem: Sem s' C t') x
  (EB: eq_below (MERGE D N x) (MERGE G M) s s'), 
  t x = t' x.
Proof.
intros.
eapply (iota_Sem _ _ (iota C F) (eq_refl _)). apply HRSem. apply HSem. apply EB.
Qed. 

(*Result called RSem_FPC_U1 in mkUProg:
  Rifle execution yields terminal tag component that respects static post-type*)
Lemma Lemma6: forall C Gamma Delta (tw:FPCT Gamma C Delta) s G M t D N F
(Hyp: JoinSem _ SVar_eq (s,G,M) (iota C F) (t,D,N))
(HF: ssqsub (zeval SVar M (allPCs F)) (Gamma pcVar))
(HG: SQSUB (MERGE G M) (FPCCTXT2CTXT Gamma)),  
SQSUB (MERGE D N) (FPCCTXT2CTXT Delta). 
Proof. 
intros C Gamma Delta tw. 
induction tw; intros; subst; simpl in *. 
(*Skip*)
inversion Hyp; subst; simpl; clear Hyp. 
  apply HG.
(*Assign*)
inversion Hyp; subst; simpl; clear Hyp.
  inversion H2; subst; clear H2. 
  inversion H4; subst; clear H4. 
  intros z; simpl.
  rewrite FPCCTXT2CTXT_upd_progVar.
  unfold FPCCTXT2CTXT, MERGE; simpl.
  remember (pvar_eq x z) as b; destruct b; subst; simpl.
    rewrite supd_Same. repeat rewrite M_upd_Same. rewrite bbot_id.
    apply lub_leq. apply lubright. apply GEval_mono. eapply leq_snart. apply HG. apply MERGE_L.
    rewrite zeval_app. 
    apply lub_leq. apply lubleft. apply HF. 
        apply lubright. rewrite zeval_GEval. apply GEval_mono. eapply leq_snart. apply HG. apply MERGE_R.
  rewrite supd_Diff; try congruence. repeat rewrite M_upd_Diff; try assumption. apply HG.
(*Comp*)
inversion Hyp; subst; clear Hyp. destruct r as [[r K] P].
assert (Z:= JoinSem_iota_allPCs _ _ _ _ _ _ _ _ H2). 
assert (IH1 := IHtw1 _ _ _ _ _ _ _ H2 HF HG). clear IHtw1.
assert (ZZ: ssqsub (zeval SVar P (allPCs F)) (G1 pcVar)).
  rewrite <- Z. 
  apply FPCT_FPC in tw1. rewrite <- (FPC_Lemma1 _ _ _ tw1). assumption.
assert (IH2 := IHtw2 _ _ _ _ _ _ _ H4 ZZ IH1). clear IHtw2.
apply IH2.
(*Ite*)
remember (fresh F) as Fr. destruct Fr as [n Fn].
inversion Hyp; subst; clear Hyp.
inversion H2; subst; clear H2.
rename G into Gamma. rename G1 into Gamma1. rename G2 into Gamma2. rename G0 into G.
inversion H4; subst; clear H4; simpl in *.
  (*True*) clear IHtw2.
     inversion H6; subst; clear H6. destruct r as [[r K] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
     assert (IH: SQSUB (MERGE K P) (FPCCTXT2CTXT Gamma1)).
        apply (IHtw1 _ _ _ _ _ _ _ H2); clear IHtw1 tw1 H2.
        rewrite fpcupd_Same. simpl.
          rewrite supd_Same. 
          apply lub_leq. apply lub_leq.  apply lubright. unfold lab_eval. simpl. rewrite id_bbot.
             eapply GEval_mono. eapply leq_snart. apply HG. apply MERGE_L.
          rewrite zeval_app. 
            apply lub_leq. apply lubleft. apply HF. 
            apply lubright. rewrite zeval_GEval. apply GEval_mono. eapply leq_snart. apply HG. apply MERGE_R.
      
        rewrite zeval_app. 
          rewrite (allPCs_eq _ M n).
            apply (leq_trans HF). apply lub_leqleft.
            intros. rewrite supd_Diff; try congruence. trivial.
        rewrite FPCCTXT2CTXT_upd_pcVar. intros z.
              unfold MERGE. rewrite supd_Diff; try discriminate. apply HG.
     assert (NN: In n (n::F)). left; trivial.
     assert (Pn := iota_ControlLabelsPreserved C1 (n::F) _ (eq_refl _) _ _ _ _ _ _ H2 n NN). rewrite supd_Same in Pn.
     clear IHtw1 tw1 H2. rewrite FPCCTXT2CTXT_upd_pcVar.
     intros z. unfold MERGE. 
       apply lub_leq. eapply @leq_trans with (y:=(MERGE K P) z). apply MERGE_L.
            apply (leq_trans (IH z)). unfold FPCCTXT2CTXT. apply FLUB_leqleft.
       remember (in_dec pvar_eq z (modVarsL C2)) as b; destruct b; subst; simpl in *.
         assert (Z: ssqsub (llub (Gamma pcVar) (GEval (fun x : PVar => Gamma (progVar x)) E)) (Gamma2 (progVar z))).
            apply FPCT_FPC in tw2. assert (MC2: modVars C2 z = true). apply modVarsChar. apply i.
                  assert (ZZ:= FPC_Mod _ _ _ _ tw2 MC2). rewrite fpcupd_Same in ZZ. apply ZZ.
         destruct (lub_leq_elim _ _ _ Z) as [Z1 Z2]. clear Z.
            apply lub_leq. rewrite <- Pn. 
                apply lub_leq. unfold lab_eval. simpl. rewrite id_bbot.
                  eapply @leq_trans with (y:= GEval (MERGE G M) E).
                    apply GEval_mono. intros y. apply lub_leqleft.
                  eapply leq_trans. apply GEval_mono. apply HG.
                  apply (leq_trans Z2). unfold FPCCTXT2CTXT; simpl. apply FLUB_leqright.
                rewrite zeval_app.
                apply lub_leq. apply (leq_trans HF). apply (leq_trans Z1). unfold FPCCTXT2CTXT, FLUB; simpl. apply lub_leqright.
                  rewrite zeval_GEval. eapply @leq_trans with (y:= GEval (MERGE G M) E).
                    apply GEval_mono. intros y. apply lub_leqright.
                  eapply leq_trans. apply GEval_mono. apply HG.
                  apply (leq_trans Z2). unfold FPCCTXT2CTXT; simpl. apply FLUB_leqright.
           eapply @leq_trans with (y:=(MERGE K P) z). apply MERGE_R.
              apply (leq_trans (IH z)). unfold FPCCTXT2CTXT; simpl. apply FLUB_leqleft.
       eapply @leq_trans with (y:=(MERGE K P) z). apply MERGE_R.
              apply (leq_trans (IH z)). unfold FPCCTXT2CTXT; simpl. apply FLUB_leqleft.
  (*False*) clear IHtw1.
     inversion H6; subst; clear H6. destruct r as [[r K] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
     assert (IH: SQSUB (MERGE K P) (FPCCTXT2CTXT Gamma2)).
        apply (IHtw2 _ _ _ _ _ _ _ H2); clear IHtw2 tw2 H2.
        rewrite fpcupd_Same. simpl.
          rewrite supd_Same. 
          apply lub_leq. apply lub_leq.  apply lubright. unfold lab_eval. simpl. rewrite id_bbot.
             eapply GEval_mono. eapply leq_snart. apply HG. apply MERGE_L.
          rewrite zeval_app. 
            apply lub_leq. apply lubleft. apply HF. 
            apply lubright. rewrite zeval_GEval. apply GEval_mono. eapply leq_snart. apply HG. apply MERGE_R.
      
        rewrite zeval_app. 
          rewrite (allPCs_eq _ M n).
            apply (leq_trans HF). apply lub_leqleft.
            intros. rewrite supd_Diff. trivial. intros Z; inversion Z; subst. apply (H H0).
            intros Z. apply (Fn Z).
        intros z. rewrite FPCCTXT2CTXT_upd_pcVar.
              unfold MERGE. rewrite supd_Diff; try discriminate. apply HG.
     assert (NN: In n (n::F)). left; trivial.
     assert (Pn := iota_ControlLabelsPreserved C2 (n::F) _ (eq_refl _) _ _ _ _ _ _ H2 n NN). rewrite supd_Same in Pn.
     clear IHtw2 tw2 H2. rewrite FPCCTXT2CTXT_upd_pcVar.
     intros z. unfold MERGE. 
       apply lub_leq. eapply @leq_trans with (y:=(MERGE K P) z). apply MERGE_L.
            apply (leq_trans (IH z)). unfold FPCCTXT2CTXT; simpl. apply FLUB_leqright.
       remember (in_dec pvar_eq z (modVarsL C1)) as b; destruct b; subst; simpl in *.
         assert (Z: ssqsub (llub (Gamma pcVar) (GEval (fun x : PVar => Gamma (progVar x)) E)) (Gamma1 (progVar z))).
            apply FPCT_FPC in tw1. assert (MC2: modVars C1 z = true). apply modVarsChar. apply i.
                  assert (ZZ:= FPC_Mod _ _ _ _ tw1 MC2). rewrite fpcupd_Same in ZZ. apply ZZ.
         destruct (lub_leq_elim _ _ _ Z) as [Z1 Z2]. clear Z.
            apply lub_leq. rewrite <- Pn. 
                apply lub_leq. unfold lab_eval. simpl. rewrite id_bbot.
                  eapply @leq_trans with (y:= GEval (MERGE G M) E).
                    apply GEval_mono. intros y. apply lub_leqleft.
                  eapply leq_trans. apply GEval_mono. apply HG.
                  apply (leq_trans Z2). unfold FPCCTXT2CTXT; simpl. apply FLUB_leqleft.
                rewrite zeval_app.
                apply lub_leq. apply (leq_trans HF). apply (leq_trans Z1). unfold FPCCTXT2CTXT; simpl. apply FLUB_leqleft.
                  rewrite zeval_GEval. eapply @leq_trans with (y:= GEval (MERGE G M) E).
                    apply GEval_mono. apply MERGE_R.
                  eapply leq_trans. apply GEval_mono. apply HG.
                  apply (leq_trans Z2). unfold FPCCTXT2CTXT; simpl. apply FLUB_leqleft.
           eapply @leq_trans with (y:=(MERGE K P) z). apply MERGE_R.
              apply (leq_trans (IH z)). unfold FPCCTXT2CTXT. simpl. apply FLUB_leqright.
       eapply @leq_trans with (y:=(MERGE K P) z). apply MERGE_R.
              apply (leq_trans (IH z)). unfold FPCCTXT2CTXT. simpl. apply FLUB_leqright.

(*While*)
rename G into Gamma. rename G0 into G.
rewrite FPCCTXT2CTXT_upd_pcVar.
clear PreBelowN.
assert (PostPre: FSQSUB (post N) (pre N)).
  rewrite PreAboveN. rewrite PrePost.
  intros z. destruct z. rewrite fpcupd_Same. apply lubleft. apply FLUB_leqleft. 
  rewrite fpcupd_Diff; try discriminate. apply FLUB_leqleft.
  trivial.
assert (PostPre': SQSUB (FPCCTXT2CTXT (post N)) (FPCCTXT2CTXT (pre N))).
  intros z. apply PostPre.
assert (HGN : FSQSUB Gamma (pre N)).
  clear Hyp. 
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. 
       simpl. eapply @leq_trans with (y:=(lub (post N pcVar) (Gamma pcVar))). apply lub_leqright.
         apply lub_leqleft. 
  rewrite fpcupd_Diff; try discriminate. apply FLUB_leqright.
  apply Le.le_refl.
intros x. unfold MERGE; simpl.
remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b; subst; simpl in *.
Focus 2.
(*~ In x modvars C*) clear H0.
  rewrite <- (iota_noMod (While E C) F _ (eq_refl _) _ _ _ _ _ _ Hyp x). 
    remember (fresh F) as Fr. destruct Fr as [m Fm].
    inversion Hyp; subst; simpl; clear Hyp.  destruct r as [[r K] P].
    inversion H2; subst; clear H2. rename r into s. rename K into G. 
    inversion H4; subst; clear H4.  destruct r as [[r K] P].
    destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
    remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b. exfalso. apply (n i).
    clear Heqb Heqb0 n0.
    destruct (JSem1 _ _  _ _ _ _ _ _ _ H2 x) as [_ RS2]; clear H2.
    simpl. rewrite modVars_iota. rewrite modVars_mkJoins.
          apply modVarsChar' in n. rewrite n. reflexivity.
    rewrite <- RS2. clear RS2.
    eapply (leq_trans (HG x)). apply HGN.
  rewrite modVars_iota. simpl. apply modVarsChar'. apply n.
(*In x modVars C*)
assert (HypCN := H0 N). clear H0. 
remember (fresh F) as Fr. destruct Fr as [n Fn].
inversion Hyp; subst; clear Hyp.
inversion H2; subst; clear H2.
inversion H4; subst; clear H4. destruct r as [[r K] P].
destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
remember (zupd _ SVar_eq M (cond_svar n)
         (llub (lab_eval G (E :: nil))
            (zeval SVar M (allPCs F ++ mkSVars E)))) as MM.
assert (HSMM: ssqsub (zeval SVar MM (allPCs (n::F))) (pre N pcVar)).
  clear H2 HypCN. subst. simpl. rewrite supd_Same.
  assert (Z: ssqsub (Gamma pcVar) (pre N pcVar)). apply HGN.
  assert (ZZ: ssqsub (GEval (MERGE G M) E) (pre N pcVar)).
    rewrite (PreAboveN N). rewrite PrePost. rewrite fpcupd_Same.
      eapply leq_trans. apply GEval_mono. apply HG.
      apply lubright. unfold FPCCTXT2CTXT, FLUB. simpl. rewrite GEval_sep'. apply lub_leqright.
    trivial.
  apply lub_leq. 
    apply lub_leq. unfold lab_eval. simpl. rewrite id_bbot. eapply (leq_snart ZZ). apply GEval_mono. apply MERGE_L.
    rewrite zeval_app. 
     apply lub_leq. eapply (leq_trans HF). apply Z.
     rewrite zeval_GEval. eapply (leq_snart ZZ). apply GEval_mono. apply MERGE_R.
   rewrite <- zeval_allPCs_supd. eapply (leq_trans HF). apply Z. intros cc. apply (Fn cc).
assert (HMM: SQSUB (MERGE G MM) (FPCCTXT2CTXT (pre N))).
  clear HypCN H2. 
    rewrite (PreAboveN N). rewrite PrePost. rewrite FPCCTXT2CTXT_upd_pcVar.
    subst. intros z. unfold MERGE, FLUB, FPCCTXT2CTXT; simpl. rewrite supd_Diff; try discriminate.
    eapply leq_snart. apply lubright. apply HG. apply leq_refl.
  trivial.
assert (HpcVarN: ssqsub (pre N pcVar) (FPCCTXT2CTXT (pre N) x)). 
  eapply leq_snart. apply PostPre.
  eapply FPC_Mod'. apply FPCT_FPC. apply HypC. apply modVarsChar. apply i.
clear HeqMM HG HF. clear M HypC PreZ.
remember ((s, G, MM)) as ST. remember (r, K, P) as TT.
remember (JWhile SVar E
          (JComp SVar (iota C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JJoin SVar (cond_svar n) (E :: nil)
                   (cond_svar n :: allPCs F ++ mkSVars E))))) as PP.
  generalize dependent P. generalize dependent K. generalize dependent r.
  generalize dependent MM. generalize dependent G. generalize dependent s.
  induction H2; intros; try inversion HeqPP; subst; clear HeqPP; simpl in *.
 (*WhileTrue*)
    clear IHJoinSem1. destruct r as [[r KK] Q].
    inversion H2_; subst; clear H2_. destruct r1 as [[u1 G1] M1].
    inversion H5; subst; clear H5. destruct r1 as [[u2 G2] M2]. 
    inversion H7; subst; clear H7.
    destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
    assert (IH := IHJoinSem2 (eq_refl _) _ _ _ (eq_refl _)). clear IHJoinSem2 H2_0.
    assert (HC:= HypCN _ _ _ _ _ _ _ H3 HSMM HMM). clear HypCN.
    assert (Pn:= iota_ControlLabelsPreserved C (n::F) _ (eq_refl _) _ _ _ _ _ _ H3).
    assert (AA:= JoinSem_iota_allPCs _ _ _ _ _ _ _ _ H3). simpl in AA.
    assert (ZZ:= iota_Mod _ _ _ (eq_refl _) _ _ _ _ _ _ H3). clear H3.
    assert (YY: llub (M1 (cond_svar n)) (M1 (var_svar (progVar x))) = M1 (var_svar (progVar x))).
       apply leq_lubright. apply ZZ. rewrite modVars_iota. apply modVarsChar. apply i. left; trivial.
    eapply IH; clear IH. 
      simpl. rewrite supd_Same.
        apply lub_leq. 
          apply lub_leq. rewrite PreAboveN. rewrite PrePost.  rewrite fpcupd_Same.  
              apply lubright. unfold FLUB. simpl. rewrite GEval_sep'. apply lubleft.
              rewrite lab_eval_GEval. apply GEval_mono. eapply leq_snart. apply HC. intros z. apply MERGE_L.
            trivial.
       apply lub_leq. rewrite <- Pn. 
          eapply leq_snart. apply HSMM. apply lub_leqleft. left; trivial.
       rewrite zeval_app.
         apply lub_leq. rewrite (allPCs_pcVar_eq _ M1).
             eapply leq_snart. apply HSMM. rewrite AA. clear AA. apply lubright.
             apply leq_refl. 
         intros z Hz. destruct z. destruct v. exfalso. apply Hz. trivial.
              remember (in_dec pvar_eq p (modVarsL C)) as b; destruct b; simpl.
                apply leq_lubright. apply ZZ. 
                         rewrite modVars_iota. apply modVarsChar. apply i0.
                       left. trivial.
              trivial.
           trivial.
       rewrite PreAboveN. rewrite PrePost. rewrite fpcupd_Same. rewrite zeval_GEval.
         apply lubright. apply GEval_mono. intros z.
          unfold FLUB; simpl. apply lubleft.
            eapply leq_snart. apply HC. apply lubright. 
            remember (in_dec pvar_eq z (modVarsL C)) as b; destruct b; simpl.
               unfold llub. rewrite leq_lubright. apply leq_refl.
               apply ZZ. rewrite modVars_iota. apply modVarsChar. apply i0.
                       left. trivial.
            apply leq_refl.
          trivial.
     rewrite <- zeval_allPCs_supd. eapply leq_snart. apply HSMM. rewrite AA.
         rewrite (allPCs_pcVar_eq _ M1). apply lub_leqright.
         intros z Hz. destruct z. destruct v. exfalso. apply Hz. trivial.
              remember (in_dec pvar_eq p (modVarsL C)) as b; destruct b; simpl.
                apply leq_lubright. apply ZZ. 
                         rewrite modVars_iota. apply modVarsChar. apply i0.
                       left. trivial.
              trivial.
           trivial.
     intros II. apply (Fn II).
  intros z. 
    rewrite PreAboveN.
      rewrite PrePost. rewrite FPCCTXT2CTXT_upd_pcVar.  
        unfold MERGE, FLUB; simpl. apply lubleft. eapply leq_snart. apply HC. 
        apply lub_leq. apply lub_leqleft.
      rewrite supd_Diff; try discriminate.
        remember (in_dec pvar_eq z (modVarsL C)) as b; destruct b; simpl. 
           unfold llub. rewrite leq_lubright. apply lub_leqright.
          apply ZZ. rewrite modVars_iota. apply modVarsChar. apply i0.
                       left. trivial.
        apply lub_leqright.
    trivial.
  reflexivity.
(*False*)
  inversion HeqTT; subst; clear HeqTT.
  clear HypCN.
  apply lub_leq. eapply leq_snart. apply HMM. apply lub_leqleft.
  remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b.
    apply lub_leq. eapply leq_snart. apply HpcVarN. eapply leq_snart. apply HSMM.
      apply lub_leqleft.
    eapply leq_snart. apply HMM. apply lub_leqright.
  eapply leq_snart. apply HMM. apply lub_leqright.
Qed.  

(*A variant of Lemma 6 (called RSem_FPC_U2 in mkUProg) that includes a guarantee on zeval  N F*)
Lemma Iota_FPC_U2: forall C Gamma Delta (tw:FPCT Gamma C Delta) s G M t D N F
(Hyp: JoinSem _ SVar_eq (s,G,M) (iota C F) (t,D,N))
(HF: ssqsub (zeval SVar M (allPCs F)) (Gamma pcVar))
(HG: SQSUB (MERGE G M) (FPCCTXT2CTXT Gamma)),  
SQSUB (MERGE D N) (FPCCTXT2CTXT Delta) /\
ssqsub (zeval SVar N (allPCs F)) (Delta pcVar). 
Proof. 
intros.
split. eapply Lemma6. apply tw. apply Hyp. apply HF. apply HG.
rewrite <- (JoinSem_iota_allPCs _ _ _ _ _ _ _ _ Hyp).
apply FPCT_FPC in tw. apply FPC_Lemma1 in tw. rewrite <- tw. apply HF. 
Qed.

(*Next: comparison between iota and theta: file IotaTheta.*)

